import numpy as np

import tensorflow as tf
from tensorflow.keras.utils import Sequence
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.optimizers.legacy import Adam

import transformers
from os.path import join


class DataGenerator(Sequence):

    def __init__(self, examples, labels, batch_size, vectorizer):
        self.examples = examples
        self.labels = labels
        self.batch_size = batch_size
        self.vectorizer = vectorizer

    def __len__(self):
        return (np.ceil(len(self.examples) / self.batch_size)).astype(np.int32)

    def __getitem__(self, idx):
        batch_x_ex = self.examples[idx * self.batch_size: (idx + 1) * self.batch_size]
        input_ids, attention_masks, token_type_ids = self.vectorizer.vectorize(batch_x_ex)
        batch_y = self.labels[idx * self.batch_size: (idx + 1) * self.batch_size]
        labels = np.array(batch_y, dtype="int32")
        return [input_ids, attention_masks, token_type_ids], labels


class RelationModelBertTrainer:

    def __init__(self, vectorizer, batch_size, num_epochs):
        self.vectorizer = vectorizer
        self.batch_size = batch_size
        self.num_epochs = num_epochs

    def get_model(self):
        model = transformers.TFBertForSequenceClassification.from_pretrained('bert-base-german-dbmdz-uncased',
                                                                             num_labels=2, from_pt=True)
        model.resize_token_embeddings(self.vectorizer.vocab_size)

        return model

    def train_model(self, train_examples, train_labels, val_examples, val_labels, output_dir):
        model = self.get_model()

        loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        model.compile(loss=loss, optimizer=Adam(0.00001), metrics=['accuracy'])
        print(model.summary())

        early_stopping = EarlyStopping(monitor='val_loss', patience=3)

        checkpoint_model_path = join(output_dir, f'bert.h5')
        model_checkpoint = ModelCheckpoint(checkpoint_model_path, monitor='val_loss', mode='min', save_best_only=True,
                                           save_weights_only=True)

        training_generator = DataGenerator(train_examples, train_labels, self.batch_size, self.vectorizer)
        validation_generator = DataGenerator(val_examples, val_labels, self.batch_size, self.vectorizer)

        model.fit(x=training_generator,
                  steps_per_epoch=int(len(train_examples) // self.batch_size),
                  epochs=self.num_epochs,
                  validation_data=validation_generator,
                  validation_steps=int(len(val_examples) // self.batch_size),
                  callbacks=[early_stopping, model_checkpoint])

        model.load_weights(checkpoint_model_path)
        best_model_path = join(output_dir, 'bert')
        model.save_pretrained(best_model_path)



from dataclasses import dataclass
from typing import Optional

from proquo.core.Reference import Reference
from proquo.core.Quote import Quote


@dataclass
class QuoteRef:
    quote: Quote
    reference: Optional[Reference]
    text: str
    pred: float = 0



from collections import OrderedDict
from typing import List, Optional
from keras_preprocessing.sequence import pad_sequences
from tensorflow_text.tools.wordpiece_vocab import bert_vocab_from_dataset as bert_vocab
from tensorflow_text import BertTokenizer
import tensorflow as tf
import re


class RelationVectorizerLstm:
    max_length: int
    vocab_size: int
    tokenizer: BertTokenizer

    def __init__(self, lookup_table, max_length, lower_case):
        self.max_length = max_length
        self.vocab_size = int(lookup_table.size())
        self.tokenizer = BertTokenizer(lookup_table, lower_case=lower_case)

    @classmethod
    def from_raw(cls, fit_data, max_length, lower_case=True, vocab_save_path: Optional[str] = None):
        fit_data_mod = []

        for elem in fit_data:
            elem = re.sub(r'<REF>|<OREF>|<Q>|</Q>', '', elem, flags=re.DOTALL)
            fit_data_mod.append(elem)

        dataset = tf.data.Dataset.from_tensor_slices(fit_data_mod)

        bert_tokenizer_params = dict(lower_case=lower_case)
        reserved_tokens = ["[PAD]", "[UNK]", "<REF>", "<OREF>", "<Q>", "</Q>"]

        bert_vocab_args = dict(
            # The target vocabulary size
            vocab_size=8000,
            # Reserved tokens that must be included in the vocabulary
            reserved_tokens=reserved_tokens,
            # Arguments for `text.BertTokenizer`
            bert_tokenizer_params=bert_tokenizer_params,
            # Arguments for `wordpiece_vocab.wordpiece_tokenizer_learner_lib.learn`
            learn_params={},
        )

        vocab = bert_vocab.bert_vocab_from_dataset(
            dataset.batch(1000).prefetch(2),
            **bert_vocab_args
        )

        if vocab_save_path:
            with open(vocab_save_path, 'w', encoding='utf-8') as f:
                for token in vocab:
                    print(token, file=f)

        lookup_table = tf.lookup.StaticVocabularyTable(
            num_oov_buckets=1,
            initializer=tf.lookup.KeyValueTensorInitializer(
                keys=vocab,
                values=tf.range(len(vocab), dtype=tf.int64)))

        return cls(lookup_table, max_length, lower_case)

    @classmethod
    def from_vocab_file(cls, vocab_file_path, max_length, lower_case):
        lookup_table = tf.lookup.StaticVocabularyTable(
            num_oov_buckets=1,
            initializer=tf.lookup.TextFileInitializer(
                filename=vocab_file_path,
                key_dtype=tf.string,
                key_index=tf.lookup.TextFileIndex.WHOLE_LINE,
                value_dtype=tf.int64,
                value_index=tf.lookup.TextFileIndex.LINE_NUMBER))

        return cls(lookup_table, max_length, lower_case)

    def vectorize(self, input_list: List[str]):
        special_tokens = {'<REF>': 2, '<OREF>': 3, '<Q>': 4, '</Q>': 5}
        special_token_positions = []

        for sent in input_list:
            parts = re.split(r'(<REF>|<OREF>|<Q>|</Q>)', sent)
            pos_map = OrderedDict()
            offset = 0
            special_offset = 0
            for part in parts:
                if part in special_tokens:
                    start = offset
                    token_id = special_tokens[part]
                    pos_map[start - special_offset] = token_id
                    special_offset += len(part)

                offset += len(part)

            special_token_positions.append(pos_map)

        input_list_mod = []

        for elem in input_list:
            elem = re.sub(r'<REF>|<OREF>|<Q>|</Q>', '', elem, flags=re.DOTALL)
            input_list_mod.append(elem)

        pieces, starts, ends = self.tokenizer.tokenize_with_offsets(input_list_mod)
        pieces_merged = pieces.merge_dims(-2, -1)
        starts_merged = starts.merge_dims(-2, -1)
        ends_merged = ends.merge_dims(-2, -1)

        pieces_list = pieces_merged.to_list()
        starts_list = starts_merged.to_list()
        ends_list = ends_merged.to_list()

        for piece, starts, ends, pos_map in zip(pieces_list, starts_list, ends_list, special_token_positions):
            for token_pos, token_id in reversed(pos_map.items()):
                if token_pos <= starts[0]:
                    piece.insert(0, token_id)
                elif token_pos > ends[-1]:
                    piece.append(token_id)
                else:
                    for list_pos, end_pos in enumerate(ends):
                        if token_pos <= end_pos:
                            piece.insert(list_pos, token_id)
                            break

        sent_ints_padded = pad_sequences(pieces_list, self.max_length)
        return sent_ints_padded



from proquo.model.relation.RelationModelLstmTrainer import RelationModelLstmTrainer


def train(train_file_path, val_file_path, output_folder_path):

    x_train = []
    y_train = []

    with open(train_file_path, 'r', encoding='utf-8') as train_file:
        for line in train_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                x_train.append(parts[0])
                y_train.append(int(parts[1]))
            else:
                print(f'wrong count: {line}')

    x_val = []
    y_val = []

    with open(val_file_path, 'r', encoding='utf-8') as val_file:
        for line in val_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                x_val.append(parts[0])
                y_val.append(int(parts[1]))
            else:
                print(f'wrong count: {line}')

    model = RelationModelLstmTrainer(200, True, 64, 64, 0.3, 128, 5)
    model.train_model(x_train, y_train, x_val, y_val, output_folder_path)



from argparse import ArgumentParser
import sys
import logging

from os.path import join, isfile, splitext, basename, isdir, exists
from os import listdir
from datetime import datetime
from pathlib import Path
from typing import List

from quid.core.Quid import Quid
from quid.match.MatchSpan import MatchSpan

from proquo.cli.Helper import get_quid_matches_mp
from proquo.cli.OptionValueCheckAction import OptionValueCheckAction
from proquo.match.MatchRef import MatchRef
from proquo.core.ProQuo import ProQuo
from proquo.core import Helper
from proquo.model.reference.ReferenceModelTrainer import ReferenceModelTrainer
from proquo.model.relation.RelationVectorizerBert import RelationVectorizerBert
import transformers

from proquo.model.reference.ReferenceVectorizer import ReferenceVectorizer
from proquo.testing.reference import TestReference
from proquo.testing.relation import TestRelationBert
from proquo.training.reference import TrainReference
from proquo.training.relation import TrainRelationBert
import re
import json

# only import if tensorflow_text is found. This is hardly ever needed and tensorflow_text is not always easy to install.
try:
    from proquo.training.relation import TrainRelationLstm
    from proquo.testing.relation import TestRelationLstm
except ModuleNotFoundError:
    pass


def __json_encoder_proquo(obj):
    if isinstance(obj, MatchRef):
        result_dict = obj.__dict__

        if not result_dict['reference']:
            del result_dict['reference']

        return result_dict

    if isinstance(obj, MatchSpan):
        result_dict = obj.__dict__

        if not result_dict['text']:
            del result_dict['text']

        return result_dict

    return obj.__dict__


def __process_file(pro_quo, filename, source_file_content, target_file_content, quid_matches_all, quid_matches_long,
                   output_folder_path, source_text_parallel_print, parallel_print_first_page, parallel_print_last_page,
                   export_text, export_ref, output_type, csv_sep):

    logging.info(f'Processing {filename}')
    short_matches: List[MatchRef] = pro_quo.compare(source_file_content, target_file_content, quid_matches_all,
                                                    source_text_parallel_print, parallel_print_first_page,
                                                    parallel_print_last_page)
    all_matches = short_matches

    if len(quid_matches_long) > 0:
        all_matches.extend(quid_matches_long)
        all_matches = Helper.remove_overlapping_matches(all_matches, target_file_content)
        all_matches.sort(key=lambda x: x.target_span.start, reverse=False)

    if not export_text:
        for match in all_matches:
            match.source_span.text = ''
            match.target_span.text = ''

    if not export_ref:
        for match in all_matches:
            match.reference = None

    if output_type == 'json':
        result = json.dumps(all_matches, default=__json_encoder_proquo)
        file_ending = 'json'
    elif output_type == 'csv':
        result = f'sstart{csv_sep}send{csv_sep}tstart{csv_sep}tend{csv_sep}stext{csv_sep}ttext'

        if export_ref:
            result += f'{csv_sep}rstart{csv_sep}rend{csv_sep}rtext'

        for match in all_matches:
            source_span = match.source_span
            target_span = match.target_span

            result += f'\n{source_span.start}{csv_sep}{source_span.end}' \
                      f'{csv_sep}{target_span.start}{csv_sep}{target_span.end}'

            if export_text:
                source_span_text = re.sub(rf'[{csv_sep}\n]', ' ', source_span.text)
                target_span_text = re.sub(rf'[{csv_sep}\n]', ' ', target_span.text)
                result += f'{csv_sep}{source_span_text}{csv_sep}{target_span_text}'

            if export_ref and match.reference:
                ref_text = re.sub(rf'[{csv_sep}\n]', ' ', match.reference.text)
                result += f'{csv_sep}{match.reference.start}{csv_sep}{match.reference.end}{csv_sep}{ref_text}'
            elif export_ref:
                result += f'{csv_sep}{csv_sep}{csv_sep}'

        file_ending = 'csv'
    else:
        result = ''

        for match in all_matches:
            result += f'\n{match.source_span.start}\t{match.source_span.end}' \
                      f'\t{match.target_span.start}\t{match.target_span.end}'

            if export_text:
                result += f'\t{match.source_span.text}\t{match.target_span.text}'

            if export_ref and match.reference:
                result += f'\t{match.reference.start}\t{match.reference.end}\t{match.reference.text}'

        result = result.strip()
        file_ending = 'txt'

    if output_folder_path:
        filename = f'{filename}.{file_ending}'

        with open(join(output_folder_path, filename), 'w', encoding='utf-8') as output_file:
            output_file.write(result)
    else:
        print('Results:')
        print(result)


def __train_reference(train_file_path, val_file_path, output_path):
    TrainReference.train(train_file_path, val_file_path, output_path)


def __train_relation(train_file_path, val_file_path, output_path, arch_type):
    if arch_type == 'bert':
        TrainRelationBert.train(train_file_path, val_file_path, output_path)
    elif arch_type == 'lstm':
        TrainRelationLstm.train(train_file_path, val_file_path, output_path)


def __test_reference(test_file_path, vocab_file_path, model_file_path):
    TestReference.test(test_file_path, vocab_file_path, model_file_path)


def __test_relation_lstm(test_file_path, vocab_file_path, model_file_path):
    TestRelationLstm.test(test_file_path, vocab_file_path, model_file_path)


def __test_relation_bert(test_file_path, tokenizer_folder_path, model_folder_path):
    TestRelationBert.test(test_file_path, tokenizer_folder_path, model_folder_path)


def __run_compare(source_file_path, target_path, ref_vocab_file_path, ref_model_file_path, rel_tokenizer_folder_path,
                  rel_model_folder_path, output_folder_path, parallel_print_files,
                  parallel_print_first_page, parallel_print_last_page, export_text, export_ref, output_type, csv_sep,
                  open_quote, close_quote, include_long_matches_in_result, max_num_processes):

    reference_vectorizer = ReferenceVectorizer.from_vocab_file(ref_vocab_file_path, 25, True)
    reference_model_trainer = ReferenceModelTrainer(25, True, 32, 32, 0.2, 512, 10)
    reference_model = reference_model_trainer.get_model(reference_vectorizer.max_id)
    reference_model.load_weights(ref_model_file_path)

    relation_vectorizer = RelationVectorizerBert.from_saved(200, rel_tokenizer_folder_path, True)
    relation_model = transformers.TFBertForSequenceClassification.from_pretrained(rel_model_folder_path, num_labels=2)

    with open(source_file_path, 'r', encoding='utf-8') as source_file:
        source_file_content = source_file.read().lower()

    pro_quo = ProQuo(reference_model, reference_vectorizer, relation_model, relation_vectorizer, open_quote,
                     close_quote)

    if isfile(target_path) and target_path.endswith('.txt'):
        with open(target_path, 'r', encoding='utf-8') as target_file:
            target_file_content = target_file.read()

        filename = splitext(basename(target_path))[0]

        quid_all = Quid(min_match_length=2, keep_ambiguous_matches=True)
        quid_matches_all = quid_all.compare(source_file_content, target_file_content)

        quid_matches_long = []
        if include_long_matches_in_result:
            quid_long = Quid(min_match_length=5, keep_ambiguous_matches=False)
            quid_matches_long = quid_long.compare(source_file_content, target_file_content)

        source_text_parallel_print = False

        if filename in parallel_print_files:
            source_text_parallel_print = True

        __process_file(pro_quo, filename, source_file_content, target_file_content, quid_matches_all, quid_matches_long,
                       output_folder_path, source_text_parallel_print, parallel_print_first_page,
                       parallel_print_last_page, export_text, export_ref, output_type, csv_sep)
    elif isdir(target_path):
        quid_matches_all_per_file = get_quid_matches_mp(source_file_content, target_path, max_num_processes, 2, True)

        quid_matches_long_per_file = None
        if include_long_matches_in_result:
            quid_matches_long_per_file = get_quid_matches_mp(source_file_content, target_path, max_num_processes, 5,
                                                             False)

        file_pos = 0
        for fileOrFolder in listdir(target_path):
            target_file_path = join(target_path, fileOrFolder)

            if isfile(target_file_path) and target_file_path.endswith('.txt'):
                filename = splitext(basename(target_file_path))[0]

                with open(target_file_path, 'r', encoding='utf-8') as target_file:
                    target_file_content = target_file.read()

                quid_matches_all = quid_matches_all_per_file[file_pos]

                quid_matches_long = []
                if quid_matches_long_per_file:
                    quid_matches_long = quid_matches_long_per_file[file_pos]

                source_text_parallel_print = False

                if filename in parallel_print_files:
                    source_text_parallel_print = True

                __process_file(pro_quo, filename, source_file_content, target_file_content, quid_matches_all,
                               quid_matches_long, output_folder_path, source_text_parallel_print,
                               parallel_print_first_page, parallel_print_last_page, export_text, export_ref,
                               output_type, csv_sep)

                file_pos += 1


def main(argv=None):
    proquo_description = 'ProQuo is a tool for the detection of short quotations (<= 4 words) between two texts, a' \
                         ' source text and a target text. The target text is the text quoting the source text.' \
                         ' Quotations in the target text need to be clearly marked with quotations marks.'
    train_description = 'ProQuo train allows the user to train their own models.'
    train_reference_description = 'ProQuo train reference allows the user to train their own reference model.'
    train_relation_description = 'ProQuo train relation allows the user to train their own relation model.'
    test_description = 'ProQuo test allows the user to test their trained models.'
    test_reference_description = 'ProQuo test reference allows the user to test their trained reference model.'
    test_relation_description = 'ProQuo test relation allows the user to test their trained relation model.'
    test_relation_lstm_description = 'ProQuo test relation lstm allows the user to test their trained lstm relation' \
                                     ' model.'
    test_relation_bert_description = 'ProQuo test relation lstm allows the user to test their trained bert relation' \
                                     ' model.'

    compare_description = 'ProQuo compare allows the user to find short quotations (<= 4 words) in two texts, a' \
                          ' source text and a target text. The target text is the text quoting the source text.' \
                          ' Quotations in the target text need to be clearly marked with quotations marks.'

    argument_parser = ArgumentParser(prog='proquo', description=proquo_description)

    subparsers_command = argument_parser.add_subparsers(dest='command')
    subparsers_command.required = True

    parser_train = subparsers_command.add_parser('train', help=train_description, description=train_description)
    subparsers_train_model = parser_train.add_subparsers(dest='train_model')
    subparsers_train_model.required = True

    parser_train_reference = subparsers_train_model.add_parser('reference', help=train_reference_description,
                                                               description=train_reference_description)

    parser_train_reference.add_argument('train_file_path', nargs=1, metavar='train-file-path',
                                        help='Path to the txt file containing the training examples')
    parser_train_reference.add_argument('val_file_path', nargs=1, metavar='val-file-path',
                                        help='Path to the txt file containing the validation examples')
    parser_train_reference.add_argument('output_folder_path', nargs=1, metavar='output-folder_path',
                                        help='Path to the folder for storing the output model and vocabulary')
    parser_train_reference.add_argument('--create-dated-subfolder', dest='create_dated_subfolder', default=False,
                                        action='store_true',
                                        help='Create a subfolder named with the current date to store the results')
    parser_train_reference.add_argument('--no-create-dated-subfolder', dest='create_dated_subfolder',
                                        action='store_false',
                                        help='Do not create a subfolder named with the current date to store the '
                                             'results')

    parser_train_relation = subparsers_train_model.add_parser('relation', help=train_relation_description,
                                                              description=train_relation_description)

    parser_train_relation.add_argument('train_file_path', nargs=1, metavar='train-file-path',
                                       help='Path to the txt file containing the training examples')
    parser_train_relation.add_argument('val_file_path', nargs=1, metavar='val-file-path',
                                       help='Path to the txt file containing the validation examples')
    parser_train_relation.add_argument('output_folder_path', nargs=1, metavar='output-folder-path',
                                       help='Path to the folder for storing the output model and vocabulary')
    parser_train_relation.add_argument('--create-dated-subfolder', dest='create_dated_subfolder', default=False,
                                       action='store_true',
                                       help='Create a subfolder named with the current date to store the results')
    parser_train_relation.add_argument('--no-create-dated-subfolder', dest='create_dated_subfolder',
                                       action='store_false',
                                       help='Do not create a subfolder named with the current date to store the '
                                            'results')
    parser_train_relation.add_argument('--arch', choices=['lstm', 'bert'], dest='arch_type', default='bert',
                                       help='The model architecture to train')

    parser_test = subparsers_command.add_parser('test', help=test_description, description=test_description)
    subparsers_test_model = parser_test.add_subparsers(dest='test_model')
    subparsers_test_model.required = True

    parser_test_reference = subparsers_test_model.add_parser('reference', help=test_reference_description,
                                                             description=test_reference_description)

    parser_test_reference.add_argument('test_file_path', nargs=1, metavar='test-file-path',
                                       help='Path to the txt file containing the testing examples')
    parser_test_reference.add_argument('vocab_file_path', nargs=1, metavar='vocab-file-path',
                                       help='Path to the vocab file')
    parser_test_reference.add_argument('model_file_path', nargs=1, metavar='model-file-path',
                                       help='Path to the model file')

    parser_test_relation = subparsers_test_model.add_parser('relation', help=test_relation_description,
                                                            description=test_relation_description)

    subparsers_test_relation_arch = parser_test_relation.add_subparsers(dest='test_model_relation_arch')
    subparsers_test_relation_arch.required = True

    parser_test_relation_lstm = subparsers_test_relation_arch.add_parser('lstm', help=test_relation_lstm_description,
                                                                         description=test_relation_lstm_description)

    parser_test_relation_lstm.add_argument('test_file_path', nargs=1, metavar='test-file-path',
                                           help='Path to the txt file containing the testing examples')
    parser_test_relation_lstm.add_argument('vocab_file_path', nargs=1, metavar='vocab-file-path',
                                           help='Path to the vocab file')
    parser_test_relation_lstm.add_argument('model_file_path', nargs=1, metavar='model-file-path',
                                           help='Path to the model file')

    parser_test_relation_bert = subparsers_test_relation_arch.add_parser('bert', help=test_relation_bert_description,
                                                                         description=test_relation_bert_description)

    parser_test_relation_bert.add_argument('test_file_path', nargs=1, metavar='test-file-path',
                                           help='Path to the txt file containing the testing examples')
    parser_test_relation_bert.add_argument('tokenizer_folder_path', nargs=1, metavar='tokenizer-folder-path',
                                           help='Path to the vocab file')
    parser_test_relation_bert.add_argument('model_folder_path', nargs=1, metavar='model-folder-path',
                                           help='Path to the model file')

    parser_compare = subparsers_command.add_parser('compare', help=compare_description, description=compare_description)

    parser_compare.add_argument('source_file_path', nargs=1, metavar='source-file-path',
                                help='Path to the source text file')
    parser_compare.add_argument('target_path', nargs=1, metavar='target-path',
                                help='Path to the target text file or folder')
    parser_compare.add_argument('ref_vocab_file_path', nargs=1, metavar='ref-vocab-file-path',
                                help='Path to the reference vocab text file')
    parser_compare.add_argument('ref_model_file_path', nargs=1, metavar='ref-model-file-path',
                                help='Path to the reference model file')
    parser_compare.add_argument('rel_tokenizer_folder_path', nargs=1, metavar='rel-tokenizer-folder-path',
                                help='Path to the relation tokenizer folder')
    parser_compare.add_argument('rel_model_folder_path', nargs=1, metavar='rel-model-folder-path',
                                help='Path to the relation model folder')
    parser_compare.add_argument('--quid-match-path', dest='quid_match_path',
                                help='Path to the file or folder with quid matches. If this option is not set, then'
                                     ' Quid is used to find long matches.')
    parser_compare.add_argument('--output-folder-path', dest='output_folder_path',
                                help='The output folder path. If this option is set the output will be saved to a file'
                                     ' created in the specified folder')
    parser_compare.add_argument('--create-dated-subfolder', dest='create_dated_subfolder', default=False,
                                action='store_true',
                                help='Create a subfolder named with the current date to store the results')
    parser_compare.add_argument('--no-create-dated-subfolder', dest='create_dated_subfolder',
                                action='store_false',
                                help='Do not create a subfolder named with the current date to store the results')
    parser_compare.add_argument('--parallel-print-files', dest='parallel_print_files', nargs='*', default=[],
                                help='Filenames of files which quote a parallel print edition')
    parser_compare.add_argument('--parallel-print-first-page', dest='parallel_print_first_page', default=0,
                                help='Number of the first page with parallel print')
    parser_compare.add_argument('--parallel-print-last-page', dest='parallel_print_last_page', default=0,
                                help='Number of the last page with parallel print')
    parser_compare.add_argument('--text', dest='export_text', default=True, action='store_true',
                                help='Include matched text in the returned data structure')
    parser_compare.add_argument('--no-text', dest='export_text', action='store_false',
                                help='Do not include matched text in the returned data structure')
    parser_compare.add_argument('--ref', dest='export_ref', default=False, action='store_true',
                                help='Include matched reference in the returned data structure')
    parser_compare.add_argument('--no-ref', dest='export_ref', action='store_false',
                                help='Do not include matched reference in the returned data structure')
    parser_compare.add_argument('--output-type', choices=['json', 'text', 'csv'], dest='output_type', default='json',
                                help='The output type')
    parser_compare.add_argument('--csv-sep', dest='csv_sep', type=str, help='output separator for csv'
                                                                            ' (default: \'\\t\')', default='\t')
    parser_compare.add_argument('--open-quote', dest='open_quote', type=str,
                                help='The quotation open character. If this option is not set, then the type of'
                                     ' quotation marks used in a target text is auto automatically identified.')
    parser_compare.add_argument('--close-quote', dest='close_quote', type=str,
                                help='The quotation close character. If this option is not set, then the type of'
                                     ' quotation marks used in a target text is auto automatically identified.')
    parser_compare.add_argument('--include-long-matches-in-result', dest='include_long_matches_in_result',
                                default=False, action='store_true', help='Include matches longer than 4 words in the'
                                                                         ' output')
    parser_compare.add_argument('--max-num-processes', dest="max_num_processes", action=OptionValueCheckAction,
                                default=1, type=int, help="Maximum number of processes to use for parallel processing."
                                                          "This can significantly speed up the process.")

    args = argument_parser.parse_args(argv)

    if args.command == 'train':
        if args.train_model == 'reference':
            train_file_path = args.train_file_path[0]
            val_file_path = args.val_file_path[0]
            output_folder_path = args.output_folder_path[0]
            create_dated_subfolder = args.create_dated_subfolder

            if output_folder_path:
                if not exists(output_folder_path):
                    raise Exception(f'{output_folder_path} does not exist!')

            if create_dated_subfolder:
                now = datetime.now()
                date_time_string = now.strftime('%Y_%m_%d_%H_%M_%S')
                output_folder_path = join(output_folder_path, date_time_string)
                Path(output_folder_path).mkdir(parents=True, exist_ok=True)

            __train_reference(train_file_path, val_file_path, output_folder_path)

        elif args.train_model == 'relation':
            train_file_path = args.train_file_path[0]
            val_file_path = args.val_file_path[0]
            output_folder_path = args.output_folder_path[0]
            create_dated_subfolder = args.create_dated_subfolder
            arch_type = args.arch_type

            if create_dated_subfolder:
                now = datetime.now()
                date_time_string = now.strftime('%Y_%m_%d_%H_%M_%S')
                output_folder_path = join(output_folder_path, date_time_string)
                Path(output_folder_path).mkdir(parents=True, exist_ok=True)

            __train_relation(train_file_path, val_file_path, output_folder_path, arch_type)

    elif args.command == 'test':
        if args.test_model == 'reference':
            test_file_path = args.test_file_path[0]
            vocab_file_path = args.vocab_file_path[0]
            model_file_path = args.model_file_path[0]
            __test_reference(test_file_path, vocab_file_path, model_file_path)

        elif args.test_model == 'relation':
            if args.test_model_relation_arch == 'lstm':
                test_file_path = args.test_file_path[0]
                vocab_file_path = args.vocab_file_path[0]
                model_file_path = args.model_file_path[0]
                __test_relation_lstm(test_file_path, vocab_file_path, model_file_path)

            elif args.test_model_relation_arch == 'bert':
                test_file_path = args.test_file_path[0]
                tokenizer_folder_path = args.tokenizer_folder_path[0]
                model_folder_path = args.model_folder_path[0]
                __test_relation_bert(test_file_path, tokenizer_folder_path, model_folder_path)

    elif args.command == 'compare':
        source_file_path = args.source_file_path[0]
        target_path = args.target_path[0]
        ref_vocab_file_path = args.ref_vocab_file_path[0]
        ref_model_file_path = args.ref_model_file_path[0]
        rel_tokenizer_folder_path = args.rel_tokenizer_folder_path[0]
        rel_model_folder_path = args.rel_model_folder_path[0]
        output_folder_path = args.output_folder_path
        create_dated_subfolder = args.create_dated_subfolder
        parallel_print_files = args.parallel_print_files
        parallel_print_first_page = args.parallel_print_first_page
        parallel_print_last_page = args.parallel_print_last_page
        export_text = args.export_text
        export_ref = args.export_ref
        output_type = args.output_type
        csv_sep = bytes(args.csv_sep, 'utf-8').decode('unicode_escape')
        open_quote = args.open_quote
        close_quote = args.close_quote
        include_long_matches_in_result = args.include_long_matches_in_result
        max_num_processes = args.max_num_processes

        if create_dated_subfolder:
            now = datetime.now()
            date_time_string = now.strftime('%Y_%m_%d_%H_%M_%S')
            output_folder_path = join(output_folder_path, date_time_string)
            Path(output_folder_path).mkdir(parents=True, exist_ok=True)

        __run_compare(source_file_path, target_path, ref_vocab_file_path, ref_model_file_path,
                      rel_tokenizer_folder_path, rel_model_folder_path, output_folder_path,
                      parallel_print_files, parallel_print_first_page, parallel_print_last_page, export_text,
                      export_ref, output_type, csv_sep, open_quote, close_quote, include_long_matches_in_result,
                      max_num_processes)


if __name__ == '__main__':
    sys.exit(main())



from proquo.model.relation.RelationModelLstmTrainer import RelationModelLstmTrainer
from proquo.model.relation.RelationVectorizerLstm import RelationVectorizerLstm


def test(test_file_path, vocab_file_path, model_file_path):
    examples = []
    gold_preds = []

    with open(test_file_path, 'r', encoding='utf-8') as input_file:
        for line in input_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                examples.append(parts[0])
                gold_preds.append(int(parts[1]))

    vectorizer = RelationVectorizerLstm.from_vocab_file(vocab_file_path, 200, True)
    relation_model_trainer = RelationModelLstmTrainer(200, True, 64, 64, 0.3, 128, 5)
    relation_model = relation_model_trainer.get_model(vectorizer.vocab_size)
    relation_model.load_weights(model_file_path)

    test_data = vectorizer.vectorize(examples)
    preds = list(relation_model.predict(test_data, verbose=1).ravel())

    tp_cnt = 0
    fp_cnt = 0
    tn_cnt = 0
    fn_cnt = 0

    for example, pred, gold_pred in zip(examples, preds, gold_preds):

        if pred >= 0.3:
            if gold_pred == 1:
                tp_cnt += 1
            else:
                fp_cnt += 1
                print(f'FP: {example}, {pred}')
        else:
            if gold_pred == 0:
                tn_cnt += 1
            else:
                fn_cnt += 1
                print(f'FN: {example}, {pred}')

    precision = tp_cnt / (tp_cnt + fp_cnt)
    recall = tp_cnt / (tp_cnt + fn_cnt)

    f_score = 0
    if precision + recall > 0:
        f_score = (2 * precision * recall) / (precision + recall)

    print(f'TP: {tp_cnt}, FP: {fp_cnt}, TN: {tn_cnt}, FN: {fn_cnt}, Precision: {precision}, Recall: {recall},'
          f' F-Score: {f_score}')



import re
from typing import List, Tuple
from quid.match.Match import Match
from quid.match.MatchSpan import MatchSpan

from proquo.core import Helper
from proquo.core.Quote import Quote
import tensorflow as tf

from proquo.model.linking.LinkingVectorizer import LinkingVectorizer
from transformers.modeling_tf_utils import TFPreTrainedModel

import logging

from kpcommons import Util, Footnote


# noinspection PyMethodMayBeStatic
class ProQuoLm:
    # TODO: Make these customizable
    SCORE_CUTOFF: int = 0.85
    WITHOUT_REF_SEARCH_RADIUS = 500
    SOURCE_PARALLEL_LAST_PAGE = 63
    LONG_MIN_LENGTH = 5
    LINKING_MAX_LENGTH = 200
    BERT_LINK_MIN_PROB: float = 0.5

    def __init__(self, linking_model: TFPreTrainedModel, linking_vectorizer: LinkingVectorizer, open_quote: str = None,
                 close_quote: str = None):
        """
        :param linking_model: The model
        :param linking_vectorizer: The vectorizer to use to tokenize and encode inputs
        :param open_quote: Character used for opening quotation marks
        :param close_quote: Character used for closing quotation marks
        """
        self.source_cache = {}
        self.hashes = None
        self.source_text_parallel_print = False
        self.linking_model = linking_model
        self.linking_vectorizer = linking_vectorizer
        self.open_quote = open_quote
        self.close_quote = close_quote

    def compare(self, source_text: str, target_text: str, quid_matches: List[Match]) -> List[Match]:
        """
        Compare the two input texts and return a list of matching sequences.
        :param source_text: A source text
        :param target_text: A target text
        :param quid_matches: A list of matches from Quid to be used as candidates for short matches
        :return: A list of found matches
        """
        source_text_cleaned = Helper.clean_text(source_text)
        source_text_hash = hash(source_text_cleaned)
        if source_text_hash in self.source_cache:
            self.hashes = self.source_cache[source_text_hash]
        else:
            self.hashes = Helper.init_lsh_hashes(source_text_cleaned)
            self.source_cache[source_text_hash] = self.hashes

        short_matches: List[Match] = Helper.filter_short_matches(source_text_cleaned, target_text, quid_matches,
                                                                 self.LONG_MIN_LENGTH)

        all_quotes: List[Quote] = Helper.get_quotations(target_text, self.LONG_MIN_LENGTH, self.open_quote,
                                                        self.close_quote)
        footnote_ranges: List[Tuple[int, int]] = Footnote.get_footnote_ranges_without_offset(target_text)

        main_quotes: List[Quote] = []

        for q in all_quotes:
            if not Footnote.is_range_in_ranges(q.start, q.end, footnote_ranges):
                main_quotes.append(q)

        result_matches_bert: List[Match] = self.__predict_links(main_quotes, short_matches, source_text,
                                                                source_text_cleaned, target_text)
        return result_matches_bert

    def __predict_links(self, short_quotes: List[Quote], short_matches: List[Match], source_text: str,
                        source_text_cleaned: str, target_text: str) -> List[Match]:
        """
        Takes a list of short quotes and tries to find matches in the source text.
        :param short_quotes: A list of short quotes
        :param short_matches: A list of short matches from Quid
        :param source_text: The source text
        :param target_text: The target text
        :return: A list of matches, i.e. the short quotes for which a match in the source text could be determined.
        """

        result: List[Match] = []

        for sq in short_quotes:
            if not re.search(r'\w', sq.text):
                logging.warning(f'Quote "{sq.text}" does not contain any characters!')
                continue

            match_len = len(sq.text.split())

            if match_len == 1:
                candidates = self.__search_single_word(sq, source_text_cleaned)
            else:
                candidates = self.__search_multi_word(short_matches, sq, source_text_cleaned)

            if len(candidates) == 0:
                continue

            combinations = []

            for c in candidates:
                le_source_text, le_target_text = self.__prepare_link_texts(sq, c, source_text_cleaned, target_text)
                combinations.append((le_source_text, le_target_text))

            preds = self.__predict_link(combinations)

            best_candidate = None
            best_pred = 0

            for c, pred in zip(candidates, preds):
                if pred > self.BERT_LINK_MIN_PROB:
                    if pred > best_pred:
                        best_pred = pred
                        best_candidate = c

            if best_candidate:
                quote_source_text = source_text[best_candidate[0]:best_candidate[1]]
                source_span = MatchSpan(best_candidate[0], best_candidate[1], quote_source_text)
                target_span = MatchSpan(sq.start, sq.end, sq.text)
                match = Match(source_span, target_span)
                result.append(match)

        return result

    def __prepare_link_texts(self, sq: Quote, candidate: Tuple[int, int], source_text: str, target_text: str)\
            -> Tuple[str, str]:
        """
        Prepare passages from the source and target text to be used in the linking model to determine the best match
        for a quotation.
        :param sq: A quote
        :param candidate: A list of start and end positions of candidates in the source text
        :param source_text: The source text
        :param target_text: The target text
        :return: The prepared source and target text passages
        """

        source_start = candidate[0]
        source_end = candidate[1]
        target_start = sq.start
        target_end = sq.end

        source_quote_text = source_text[source_start:source_end].replace('\n', ' ')
        source_quote_length = len(source_quote_text.split())
        source_rest_len = self.LINKING_MAX_LENGTH - source_quote_length

        target_quote_text = target_text[target_start:target_end].replace('\n', ' ')
        target_quote_length = len(target_quote_text.split())
        target_rest_len = self.LINKING_MAX_LENGTH - target_quote_length

        if source_rest_len <= 0 or target_rest_len <= 0:
            return '', ''

        source_text_before = source_text[:source_start]
        source_text_after = source_text[source_end:]

        source_text_before = re.sub(r'\[\[\[((?:.|\n)+?)]]]', ' ', source_text_before)
        source_text_after = re.sub(r'\[\[\[((?:.|\n)+?)]]]', ' ', source_text_after)

        target_text_before = target_text[:target_start]
        target_text_after = target_text[target_end:]

        target_text_before = re.sub(r'\[\[\[((?:.|\n)+?)]]]', ' ', target_text_before)
        target_text_after = re.sub(r'\[\[\[((?:.|\n)+?)]]]', ' ', target_text_after)

        # TODO: check if is in footnote?

        source_parts_before = source_text_before.split()
        source_parts_after = source_text_after.split()

        source_parts_before_count = len(source_parts_before)
        source_parts_after_count = len(source_parts_after)

        source_count_before = min(round(source_rest_len / 2), source_parts_before_count)
        source_count_after = min(source_rest_len - source_count_before, source_parts_after_count)

        source_text_before = ' '.join(source_parts_before[-source_count_before:])
        source_text_after = ' '.join(source_parts_after[:source_count_after])

        le_source_text = f'{source_text_before} <S> {source_quote_text} </S> {source_text_after}'

        target_parts_before = target_text_before.split()
        target_parts_after = target_text_after.split()

        target_parts_before_count = len(target_parts_before)
        target_parts_after_count = len(target_parts_after)

        target_count_before = min(round(target_rest_len / 2), target_parts_before_count)
        target_count_after = min(target_rest_len - target_count_before, target_parts_after_count)

        target_text_before = ' '.join(target_parts_before[-target_count_before:])
        target_text_after = ' '.join(target_parts_after[:target_count_after])

        le_target_text = f'{target_text_before[:-1]} <T> {target_quote_text} </T> {target_text_after[1:]}'

        return le_source_text, le_target_text

    def __predict_link(self, pairs: List[Tuple[str, str]]) -> List[float]:
        """
        Predict the probability of each pair of belonging together, i.e. the quote in target text passage is taken from
        the corresponding source text passage.
        :param pairs: A list of pairs of strings to be predicted.
        :return: A list probabilities with a probability for each pair.
        """

        if len(pairs) == 0:
            return []

        if len(pairs) > 1000:
            logging.warning(f'Too many pairs: {len(pairs)}')
            return []

        test_data = self.linking_vectorizer.vectorize(pairs)
        prediction = self.linking_model.predict(test_data, verbose=0)
        prediction_logits = prediction.logits
        probs = tf.nn.softmax(prediction_logits, axis=1).numpy()
        preds = [row[1] for row in probs]
        return preds

    def __search_single_word(self, sq: Quote, source_text: str) -> List[Tuple[int, int]]:
        """
        Search for the given quote in the given source text and return a list of candidate start and end character
        positions.
        :param sq: The quote to search for
        :param source_text: The source text
        :return: A list of candidate start and end character positions
        """
        # First, look for an exact match
        re_matches = Helper.strict_match(sq.text, source_text)
        strict_matches_count = len(re_matches)

        # if there are 1 or more exact matches, return them all
        if strict_matches_count > 0:
            result = []

            for re_match in re_matches:
                result.append((re_match.start(), re_match.end()))

            return result

        # if there were no exact matches, make a fuzzy search, for details, see :meth:`proquo.core.Helper.fuzzy_match`
        # for details
        fuzzy_candidates = Helper.fuzzy_match(sq.text, 0, len(source_text), self.hashes, self.SCORE_CUTOFF)
        return fuzzy_candidates

    def __search_multi_word(self, short_matches: List[Match], sq: Quote, source_text: str) -> List[Tuple[int, int]]:
        """
        Search for the given quote in the given source text and return a list of candidate start and end character
        positions.
        :param short_matches: The matches from Quid which are used as candidates
        :param sq: A quote
        :param source_text: The source text
        :return: candidate start and end character positions
        """
        # First, look for an exact match
        re_matches = Helper.strict_match(sq.text, source_text)

        # if there are 1 or more exact matches, return them all
        if len(re_matches) > 0:
            result = []

            for re_match in re_matches:
                result.append((re_match.start(), re_match.end()))

            return result

        # if there were no exact matches, we use the short matches from Quid and try to find overlaps
        candidates = []

        for short_match in short_matches:
            overlap_length = Util.calculate_overlap(short_match.target_span.start, short_match.target_span.end,
                                                    sq.start, sq.end)
            quote_length = sq.end - sq.start
            percentage = overlap_length / quote_length

            if percentage >= 0.7:
                candidates.append((short_match.source_span.start, short_match.source_span.end))

        return candidates



from dataclasses import dataclass
from quid.match.Match import Match
from proquo.core import Reference


@dataclass
class MatchRef(Match):
    reference: Reference

    def __init__(self, source_span, target_span, reference: Reference = None):
        super().__init__(source_span, target_span)
        self.reference = reference

    @classmethod
    def from_match(cls, match: Match):
        return cls(match.source_span, match.target_span)



from argparse import ArgumentParser, BooleanOptionalAction
import sys
import logging
from os.path import join, isfile, splitext, basename, isdir, exists
from os import listdir
from datetime import datetime
from pathlib import Path
from typing import List

from quid.core.Quid import Quid
from quid.match.Match import Match
from quid.match.MatchSpan import MatchSpan

from proquo.cli.OptionValueCheckAction import OptionValueCheckAction
from proquo.core import Helper
from proquo.match.MatchRef import MatchRef
from proquo.core.ProQuoLm import ProQuoLm
from proquo.model.linking.LinkingVectorizer import LinkingVectorizer
import transformers

from proquo.testing.linking import TestLinking
from proquo.training.linking import TrainLinking

import re
import json

from proquo.cli.Helper import get_quid_matches_mp


def __json_encoder_proquo(obj):
    if isinstance(obj, MatchRef):
        result_dict = obj.__dict__
        del result_dict['reference']
        return result_dict

    if isinstance(obj, MatchSpan):
        result_dict = obj.__dict__

        if not result_dict['text']:
            del result_dict['text']

        return result_dict

    return obj.__dict__


def __process_file(pro_quo_lm, filename, source_file_content, target_file_content, quid_matches_all, quid_matches_long,
                   output_folder_path, export_text, output_type, csv_sep):
    logging.info(f'Processing {filename}')
    short_matches: List[MatchRef] = pro_quo_lm.compare(source_file_content, target_file_content, quid_matches_all)
    all_matches: List[Match] = short_matches

    if len(quid_matches_long) > 0:
        all_matches.extend(quid_matches_long)
        all_matches = Helper.remove_overlapping_matches(all_matches, target_file_content)
        all_matches.sort(key=lambda x: x.target_span.start, reverse=False)

    if not export_text:
        for match in all_matches:
            match.source_span.text = ''
            match.target_span.text = ''

    if output_type == 'json':
        result = json.dumps(all_matches, default=__json_encoder_proquo)
        file_ending = 'json'
    elif output_type == 'csv':
        result = f'sstart{csv_sep}send{csv_sep}tstart{csv_sep}tend{csv_sep}stext{csv_sep}ttext'

        for match in all_matches:
            source_span = match.source_span
            target_span = match.target_span

            result += f'\n{source_span.start}{csv_sep}{source_span.end}' \
                      f'{csv_sep}{target_span.start}{csv_sep}{target_span.end}'

            if export_text:
                source_span_text = re.sub(rf'[{csv_sep}\n]', ' ', source_span.text)
                target_span_text = re.sub(rf'[{csv_sep}\n]', ' ', target_span.text)
                result += f'{csv_sep}{source_span_text}{csv_sep}{target_span_text}'

        file_ending = 'csv'
    else:
        result = ''

        for match in all_matches:
            result += f'\n{match.source_span.start}\t{match.source_span.end}' \
                      f'\t{match.target_span.start}\t{match.target_span.end}'

            if export_text:
                result += f'\t{match.source_span.text}\t{match.target_span.text}'

        result = result.strip()
        file_ending = 'txt'

    if output_folder_path:
        filename = f'{filename}.{file_ending}'

        with open(join(output_folder_path, filename), 'w', encoding='utf-8') as output_file:
            output_file.write(result)
    else:
        print('Results:')
        print(result)


def __train(train_file_path, val_file_path, batch_size, num_epochs, lower_case, model_name, output_folder_path):
    TrainLinking.train(train_file_path, val_file_path, batch_size, num_epochs, lower_case, model_name,
                       output_folder_path)


def __test(test_file_path, tokenizer_folder_path, model_folder_path, lower_case):
    TestLinking.test(test_file_path, tokenizer_folder_path, model_folder_path, lower_case)


def __run_compare(source_file_path, target_path, tokenizer_folder_path, model_folder_path, lower_case,
                  output_folder_path, export_text, output_type, csv_sep, open_quote, close_quote,
                  include_long_matches_in_result, max_num_processes):
    link_vectorizer = LinkingVectorizer.from_saved(512, tokenizer_folder_path, lower_case)
    link_model = transformers.TFBertForSequenceClassification.from_pretrained(model_folder_path, num_labels=2)

    with open(source_file_path, 'r', encoding='utf-8') as source_file:
        source_file_content = source_file.read()

    pro_quo_lm = ProQuoLm(link_model, link_vectorizer, open_quote, close_quote)

    if isfile(target_path) and target_path.endswith('.txt'):
        with open(target_path, 'r', encoding='utf-8') as target_file:
            target_file_content = target_file.read()

        filename = splitext(basename(target_path))[0]

        logging.info(f'Running Quid for {filename} to get all matches')

        quid_all = Quid(min_match_length=2, keep_ambiguous_matches=True)
        quid_matches_all = quid_all.compare(source_file_content, target_file_content)

        quid_matches_long = []
        if include_long_matches_in_result:
            logging.info(f'Running Quid for {filename} to get long matches')
            quid_long = Quid(min_match_length=5, keep_ambiguous_matches=False)
            quid_matches_long = quid_long.compare(source_file_content, target_file_content)

        __process_file(pro_quo_lm, filename, source_file_content, target_file_content, quid_matches_all,
                       quid_matches_long, output_folder_path, export_text, output_type, csv_sep)
    elif isdir(target_path):
        logging.info(f'Running Quid for all files to get all matches')
        quid_matches_all_per_file = get_quid_matches_mp(source_file_content, target_path, max_num_processes, 2, True)

        quid_matches_long_per_file = None
        if include_long_matches_in_result:
            logging.info(f'Running Quid for all files to get long matches')
            quid_matches_long_per_file = get_quid_matches_mp(source_file_content, target_path, max_num_processes, 5,
                                                             False)

        file_pos = 0
        for fileOrFolder in listdir(target_path):
            target_file_path = join(target_path, fileOrFolder)

            if isfile(target_file_path) and target_file_path.endswith('.txt'):
                filename = splitext(basename(target_file_path))[0]

                with open(target_file_path, 'r', encoding='utf-8') as target_file:
                    target_file_content = target_file.read()

                quid_matches_all = quid_matches_all_per_file[file_pos]

                quid_matches_long = []
                if quid_matches_long_per_file:
                    quid_matches_long = quid_matches_long_per_file[file_pos]

                __process_file(pro_quo_lm, filename, source_file_content, target_file_content, quid_matches_all,
                               quid_matches_long, output_folder_path, export_text, output_type, csv_sep)

                file_pos += 1


def main(argv=None):
    train_description = 'ProQuoLm train allows the user to train their own models.'
    test_description = 'ProQuoLm test allows the user to test their trained model.'
    compare_description = 'ProQuoLm compare allows the user to find short quotations (<= 4 words) in two texts, a' \
                          ' source text and a target text. The target text is the text quoting the source text.' \
                          ' Quotations in the target text need to be clearly marked with quotations marks.'

    argument_parser = ArgumentParser(prog='proquolm', description='ProQuoLm is a tool to find short quotations'
                                                                  '(<= 4 words) in texts.')

    argument_parser.add_argument('--log-level', dest='log_level', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR',
                                                                           'CRITICAL'],
                                 help='Set the logging level', default='WARNING')

    subparsers_command = argument_parser.add_subparsers(dest='command')
    subparsers_command.required = True

    parser_train = subparsers_command.add_parser('train', help=train_description, description=train_description)

    parser_train.add_argument('train_file_path', nargs=1, metavar='train-file-path',
                              help='Path to the txt file containing the training examples')
    parser_train.add_argument('val_file_path', nargs=1, metavar='val-file-path',
                              help='Path to the txt file containing the validation examples')
    parser_train.add_argument('output_folder_path', nargs=1, metavar='output-folder_path',
                              help='Path to the folder for storing the output model and vocabulary')
    parser_train.add_argument('--create-dated-subfolder', dest='create_dated_subfolder', default=False,
                              action=BooleanOptionalAction,
                              help='Create a subfolder named with the current date to store the results')
    parser_train.add_argument('--base-model-name', dest="base_model_name",
                              default="dbmdz/bert-base-german-uncased", help="The model name (default: %(default)s)")
    parser_train.add_argument('--lower-case', dest="lower_case", default=True, action=BooleanOptionalAction,
                              help="Train model on lower case text")
    parser_train.add_argument('--batch-size', dest="batch_size", default=4, type=int,
                              help="The batch size (default: %(default)d)")
    parser_train.add_argument('--num-epochs', dest="num_epochs", default=3, type=int,
                              help="The number of epochs to train for (default: %(default)d)")

    parser_test = subparsers_command.add_parser('test', help=test_description, description=test_description)

    parser_test.add_argument('test_file_path', nargs=1, metavar='test-file-path',
                             help='Path to the txt file containing the testing examples')
    parser_test.add_argument('tokenizer_folder_path', nargs=1, metavar='tokenizer-folder-path',
                             help='Path to the vocab file')
    parser_test.add_argument('model_folder_path', nargs=1, metavar='model-folder-path',
                             help='Path to the model file')
    parser_test.add_argument('--lower-case', dest="lower_case", default=True, action=BooleanOptionalAction,
                             help="Test model on lower case text")

    parser_compare = subparsers_command.add_parser('compare', help=compare_description, description=compare_description)

    parser_compare.add_argument('source_file_path', nargs=1, metavar='source-file-path',
                                help='Path to the source text file')
    parser_compare.add_argument('target_path', nargs=1, metavar='target-path',
                                help='Path to the target text file or folder')
    parser_compare.add_argument('--tokenizer', dest='tokenizer', default='fredr0id/proquolm',
                                help='Name of the tokenizer to load from Hugging Face or path to the tokenizer folder')
    parser_compare.add_argument('--model', dest='model', default='fredr0id/proquolm',
                                help='Name of the model to load from Hugging Face or path to the model folder')
    parser_compare.add_argument('--lower-case', dest="lower_case", default=True, action=BooleanOptionalAction,
                                help="Run model inference on lower case text")
    parser_compare.add_argument('--output-folder-path', dest='output_folder_path',
                                help='The output folder path. If this option is set the output will be saved to a file'
                                     ' created in the specified folder')
    parser_compare.add_argument('--create-dated-subfolder', dest='create_dated_subfolder', default=False,
                                action=BooleanOptionalAction,
                                help='Create a subfolder named with the current date to store the results')
    parser_compare.add_argument('--text', dest='export_text', default=True, action=BooleanOptionalAction,
                                help='Include matched text in the returned data structure')
    parser_compare.add_argument('--output-type', choices=['json', 'text', 'csv'], dest='output_type', default='json',
                                help='The output type')
    parser_compare.add_argument('--csv-sep', dest='csv_sep', type=str, help='output separator for csv'
                                                                            ' (default: \'\\t\')', default='\t')
    parser_compare.add_argument('--open-quote', dest='open_quote', type=str,
                                help='The quotation open character. If this option is not set, then the type of'
                                     ' quotation marks used in a target text is auto automatically identified.')
    parser_compare.add_argument('--close-quote', dest='close_quote', type=str,
                                help='The quotation close character. If this option is not set, then the type of'
                                     ' quotation marks used in a target text is auto automatically identified.')
    parser_compare.add_argument('--include-long-matches-in-result', dest='include_long_matches_in_result',
                                default=False, action='store_true', help='Include matches longer than 4 words in the'
                                                                         ' output')
    parser_compare.add_argument('--max-num-processes', dest='max_num_processes', action=OptionValueCheckAction,
                                default=1, type=int, help='Maximum number of processes to use for parallel processing.'
                                                          ' This can significantly speed up the process.')

    args = argument_parser.parse_args(argv)

    log_level = args.log_level
    logging.getLogger().setLevel(logging.getLevelName(log_level))

    if args.command == 'train':
        train_file_path = args.train_file_path[0]
        val_file_path = args.val_file_path[0]
        output_folder_path = args.output_folder_path[0]
        create_dated_subfolder = args.create_dated_subfolder
        base_model_name = args.base_model_name
        lower_case = args.lower_case
        num_epochs = args.num_epochs
        batch_size = args.batch_size

        if output_folder_path:
            if not exists(output_folder_path):
                raise Exception(f'{output_folder_path} does not exist!')

        if create_dated_subfolder:
            now = datetime.now()
            date_time_string = now.strftime('%Y_%m_%d_%H_%M_%S')
            output_folder_path = join(output_folder_path, date_time_string)
            Path(output_folder_path).mkdir(parents=True, exist_ok=True)

        __train(train_file_path, val_file_path, batch_size, num_epochs, lower_case, base_model_name, output_folder_path)

    elif args.command == 'test':
        test_file_path = args.test_file_path[0]
        tokenizer_folder_path = args.tokenizer_folder_path[0]
        model_folder_path = args.model_folder_path[0]
        lower_case = args.lower_case
        __test(test_file_path, tokenizer_folder_path, model_folder_path, lower_case)

    elif args.command == 'compare':
        source_file_path = args.source_file_path[0]
        target_path = args.target_path[0]
        tokenizer_name_or_path = args.tokenizer
        model_name_or_path = args.model
        lower_case = args.lower_case
        output_folder_path = args.output_folder_path
        create_dated_subfolder = args.create_dated_subfolder
        export_text = args.export_text
        output_type = args.output_type
        csv_sep = bytes(args.csv_sep, 'utf-8').decode('unicode_escape')
        open_quote = args.open_quote
        close_quote = args.close_quote
        include_long_matches_in_result = args.include_long_matches_in_result
        max_num_processes = args.max_num_processes

        if create_dated_subfolder:
            now = datetime.now()
            date_time_string = now.strftime('%Y_%m_%d_%H_%M_%S')
            output_folder_path = join(output_folder_path, date_time_string)
            Path(output_folder_path).mkdir(parents=True, exist_ok=True)

        __run_compare(source_file_path, target_path, tokenizer_name_or_path, model_name_or_path, lower_case,
                      output_folder_path, export_text, output_type, csv_sep, open_quote, close_quote,
                      include_long_matches_in_result, max_num_processes)


if __name__ == '__main__':
    sys.exit(main())



from typing import List
from keras_preprocessing.sequence import pad_sequences
from keras_preprocessing.text import Tokenizer


class ReferenceVectorizer:

    tokenizer: Tokenizer
    max_length: int
    max_id: int
    lower_case: bool

    def __init__(self, tokenizer, max_length, lower_case):
        self.lower_case = lower_case
        self.tokenizer = tokenizer
        self.max_id = len(self.tokenizer.word_index)
        self.max_length = max_length

    @classmethod
    def from_raw(cls, fit_data, max_length, lower_case=True):
        if lower_case:
            fit_data = [x.lower() for x in fit_data]

        tokenizer = Tokenizer(char_level=True)
        tokenizer.fit_on_texts(fit_data)
        return cls(tokenizer, max_length, lower_case)

    @classmethod
    def from_vocab_file(cls, vocab_file_path, max_length, lower_case):
        word_index = {}

        with open(vocab_file_path, 'r', encoding='utf-8') as vocab_file:
            for count, line in enumerate(vocab_file):
                word_index[line.rstrip('\n')] = count + 1

        tokenizer = Tokenizer(char_level=True)
        tokenizer.word_index = word_index
        return cls(tokenizer, max_length, lower_case)

    def vectorize(self, input_list: List[str]):
        if self.lower_case:
            input_list = [x.lower() for x in input_list]

        input_sequences = self.tokenizer.texts_to_sequences(input_list)
        input_sequences_padded = pad_sequences(input_sequences, maxlen=self.max_length)
        return input_sequences_padded



from os.path import join
from pathlib import Path
from proquo.model.relation.RelationModelBertTrainer import RelationModelBertTrainer
from proquo.model.relation.RelationVectorizerBert import RelationVectorizerBert


def train(train_file_path, val_file_path, output_folder_path):

    x_train = []
    y_train = []

    with open(train_file_path, 'r', encoding='utf-8') as train_file:
        for line in train_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                x_train.append(parts[0])
                y_train.append(int(parts[1]))
            else:
                print(f'wrong count: {line}')

    x_val = []
    y_val = []

    with open(val_file_path, 'r', encoding='utf-8') as val_file:
        for line in val_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                x_val.append(parts[0])
                y_val.append(int(parts[1]))
            else:
                print(f'wrong count: {line}')

    relation_vectorizer = RelationVectorizerBert.from_raw(200, True)

    tokenizer_dir = join(output_folder_path, 'tokenizer')
    Path(tokenizer_dir).mkdir(parents=True, exist_ok=True)
    relation_vectorizer.tokenizer.save_pretrained(tokenizer_dir)

    model = RelationModelBertTrainer(relation_vectorizer, 12, 3)
    model.train_model(x_train, y_train, x_val, y_val, output_folder_path)



import multiprocessing
from os import listdir
from os.path import join, isfile
from quid.core.Quid import Quid


def __run_quid(source_file_content, target_file_content, min_match_length, keep_ambiguous_matches,
               cached_min_length_match_positions, cached_hashes):
    quid = Quid(min_match_length=min_match_length, keep_ambiguous_matches=keep_ambiguous_matches)
    matches = quid.compare(source_file_content, target_file_content, cached_min_length_match_positions, cached_hashes)
    return matches


def get_quid_matches_mp(source_file_content, target_path, num_of_processes, min_match_length, keep_ambiguous_matches):
    quid = Quid(min_match_length=min_match_length, keep_ambiguous_matches=keep_ambiguous_matches)
    min_length_match_positions, hashes = quid.prepare_source_data(source_file_content)
    pool = multiprocessing.Pool(num_of_processes)

    results = []
    for file_or_folder in listdir(target_path):
        full_path = join(target_path, file_or_folder)

        if isfile(full_path) and full_path.endswith(".txt"):
            with open(full_path, 'r', encoding='utf-8') as target_file:
                target_file_content = target_file.read()

            result = pool.apply_async(__run_quid, args=(source_file_content, target_file_content,
                                                        min_match_length, keep_ambiguous_matches,
                                                        min_length_match_positions, hashes))
            results.append(result)

    pool.close()
    pool.join()

    result_matches = []
    for result in results:
        result_matches.append(result.get())

    return result_matches



import numpy as np

import tensorflow as tf
from tensorflow.keras.utils import Sequence
from tensorflow.keras.callbacks import ModelCheckpoint
from tensorflow.keras.optimizers.legacy import Adam

import transformers
from os.path import join
import torch


class DataGenerator(Sequence):

    def __init__(self, examples, labels, batch_size, vectorizer):
        self.examples = examples
        self.labels = labels
        self.batch_size = batch_size
        self.vectorizer = vectorizer

    def __len__(self):
        return (np.ceil(len(self.examples) / self.batch_size)).astype(np.int32)

    def __getitem__(self, idx):
        batch_x_ex = self.examples[idx * self.batch_size: (idx + 1) * self.batch_size]
        input_ids, attention_masks, token_type_ids = self.vectorizer.vectorize(batch_x_ex)
        batch_y = self.labels[idx * self.batch_size: (idx + 1) * self.batch_size]
        labels = np.array(batch_y, dtype="int32")
        return [input_ids, attention_masks, token_type_ids], labels


# TODO: merge with TrainLinking.py?
def train_model(vectorizer, model_name, batch_size, num_epochs, train_examples, train_labels, val_examples,
                val_labels, output_folder_path):
    model = transformers.TFBertForSequenceClassification.from_pretrained(model_name, num_labels=2, from_pt=True)
    model.resize_token_embeddings(vectorizer.vocab_size)

    loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
    model.compile(loss=loss, optimizer=Adam(0.00001), metrics=['accuracy'])
    print(model.summary())

    checkpoint_model_path = join(output_folder_path, f'checkpoint.h5')
    model_checkpoint = ModelCheckpoint(checkpoint_model_path, monitor='val_loss', mode='min', save_best_only=True,
                                       save_weights_only=True)

    training_generator = DataGenerator(train_examples, train_labels, batch_size, vectorizer)
    validation_generator = DataGenerator(val_examples, val_labels, batch_size, vectorizer)

    model.fit(x=training_generator,
              steps_per_epoch=int(len(train_examples) // batch_size),
              epochs=num_epochs,
              validation_data=validation_generator,
              validation_steps=int(len(val_examples) // batch_size),
              callbacks=[model_checkpoint])

    model.load_weights(checkpoint_model_path)
    best_model_path = join(output_folder_path, 'model')
    model.save_pretrained(best_model_path)



from os.path import join
from pathlib import Path
from proquo.model.linking import LinkingModelTrainer
from proquo.model.linking.LinkingVectorizer import LinkingVectorizer
import json


def train(train_file_path, val_file_path, batch_size, num_epochs, lower_case, model_name, output_folder_path):
    config = {
        'base model name': model_name,
        'lower case': lower_case,
        'num epochs': num_epochs,
        'batch size': batch_size,
    }

    with open(join(output_folder_path, 'config.json'), 'w', encoding='utf-8') as config_file:
        content = json.dumps(config)
        config_file.write(content)

    x_train = []
    y_train = []

    with open(train_file_path, 'r', encoding='utf-8') as train_file:
        for line in train_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) >= 3:
                x_train.append((parts[0], parts[1]))
                y_train.append(int(parts[2]))
            else:
                print(f'wrong count: {line}')

    x_val = []
    y_val = []

    with open(val_file_path, 'r', encoding='utf-8') as val_file:
        for line in val_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) >= 3:
                x_val.append((parts[0], parts[1]))
                y_val.append(int(parts[2]))
            else:
                print(f'wrong count: {line}')

    linking_vectorizer = LinkingVectorizer.from_raw(model_name,512, lower_case)

    tokenizer_dir = join(output_folder_path, 'tokenizer')
    Path(tokenizer_dir).mkdir(parents=True, exist_ok=True)
    linking_vectorizer.tokenizer.save_pretrained(tokenizer_dir)

    LinkingModelTrainer.train_model(linking_vectorizer, model_name, batch_size, num_epochs, x_train, y_train, x_val,
                                    y_val, output_folder_path)



from typing import List, Tuple
import transformers
import tensorflow as tf

transformers.logging.set_verbosity_error()


class LinkingVectorizer:
    max_length: int
    vocab_size: int
    tokenizer: transformers.BertTokenizer

    def __init__(self, max_length, tokenizer):
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.vocab_size = len(self.tokenizer)

    @classmethod
    def from_raw(cls, model_name, max_length, lower_case):
        tokenizer = transformers.BertTokenizer.from_pretrained(
            model_name, do_lower_case=lower_case
        )

        special_tokens_dict = {'additional_special_tokens': ['<S>', '</S>', '<T>', '</T>']}
        tokenizer.add_special_tokens(special_tokens_dict)
        return cls(max_length, tokenizer)

    @classmethod
    def from_saved(cls, max_length, tokenizer_path, lower_case):
        tokenizer = transformers.BertTokenizer.from_pretrained(
            tokenizer_path, do_lower_case=lower_case
        )

        return cls(max_length, tokenizer)

    def vectorize(self, input_list: List[Tuple[str, str]]):
        encoded = self.tokenizer.batch_encode_plus(
            input_list,
            truncation=True,
            add_special_tokens=True,
            max_length=self.max_length,
            return_attention_mask=True,
            return_token_type_ids=True,
            padding='max_length',
            return_tensors="tf",
        )

        # Convert batch of encoded features to numpy array.
        input_ids = tf.constant(encoded["input_ids"], dtype="int32")
        attention_masks = tf.constant(encoded["attention_mask"], dtype="int32")
        token_type_ids = tf.constant(encoded["token_type_ids"], dtype="int32")

        return input_ids, attention_masks, token_type_ids



import numpy as np

from tensorflow.keras.utils import Sequence
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, LSTM, Bidirectional, Dropout, Embedding
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.optimizers import Adam
from os.path import join

from proquo.model.relation.RelationVectorizerLstm import RelationVectorizerLstm
import tensorflow as tf


class DataGenerator(Sequence):

    def __init__(self, examples, labels, batch_size, vectorizer):
        self.examples = examples
        self.labels = labels
        self.batch_size = batch_size
        self.vectorizer = vectorizer

    def __len__(self):
        return (np.ceil(len(self.examples) / self.batch_size)).astype(np.int32)

    def __getitem__(self, idx):
        batch_x_ex = self.examples[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_x = self.vectorizer.vectorize(batch_x_ex)
        batch_y = self.labels[idx * self.batch_size: (idx + 1) * self.batch_size]
        return batch_x, batch_y


class RelationModelLstmTrainer:

    def __init__(self, max_sequence_length, lower_case, num_dense_units, num_lstm_units, dropout_rate,
                 batch_size, num_epochs):
        self.max_sequence_length = max_sequence_length
        self.lower_case = lower_case
        self.num_lstm_units = num_lstm_units
        self.num_dense_units = num_dense_units
        self.dropout_rate = dropout_rate
        self.batch_size = batch_size
        self.num_epochs = num_epochs

    def get_model(self, vocab_size):
        model = Sequential()

        model.add(Embedding(input_dim=vocab_size, output_dim=self.num_dense_units,
                            input_length=self.max_sequence_length, mask_zero=True))
        model.add(Dropout(self.dropout_rate))
        model.add(Bidirectional(LSTM(self.num_lstm_units, dropout=self.dropout_rate, return_sequences=False)))
        model.add(Dropout(self.dropout_rate))
        model.add(Dense(1, activation='sigmoid'))

        return model

    def train_model(self, train_examples, train_labels, val_examples, val_labels, output_dir):
        vocab_path = join(output_dir, 'vocab.txt')
        vectorizer = RelationVectorizerLstm.from_raw(train_examples, self.max_sequence_length, self.lower_case,
                                                     vocab_save_path=vocab_path)

        train_labels = np.asarray(train_labels).astype(np.float16).reshape((-1, 1))
        val_labels = np.asarray(val_labels).astype(np.float16).reshape((-1, 1))

        model = self.get_model(vectorizer.vocab_size)

        model.compile(loss='binary_crossentropy', optimizer=Adam(0.01), metrics=['accuracy'])

        early_stopping = EarlyStopping(monitor='val_loss', patience=3)
        name = f'lstm_{self.num_lstm_units}_{self.dropout_rate}'

        best_model_path = join(output_dir, name + '.h5')
        model_checkpoint = ModelCheckpoint(best_model_path, monitor='val_loss', mode='min', save_best_only=True,
                                           save_weights_only=False)

        print(model.summary())

        training_generator = DataGenerator(train_examples, train_labels, self.batch_size, vectorizer)
        validation_generator = DataGenerator(val_examples, val_labels, self.batch_size, vectorizer)

        model.fit(x=training_generator,
                  steps_per_epoch=int(len(train_examples) // self.batch_size),
                  epochs=self.num_epochs,
                  validation_data=validation_generator,
                  validation_steps=int(len(val_examples) // self.batch_size),
                  callbacks=[early_stopping, model_checkpoint],
                  workers=1, use_multiprocessing=False, max_queue_size=10, shuffle=True)



from argparse import Action


class OptionValueCheckAction(Action):

    def __call__(self, parser, namespace, values, option_string=None):

        if option_string == '--max-num-processes':
            if int(values) <= 0:
                parser.error("{0} must be greater 0".format(option_string))

        setattr(namespace, self.dest, values)


import tensorflow as tf
from proquo.model.relation.RelationVectorizerBert import RelationVectorizerBert
import transformers


def test(test_file_path, tokenizer_folder_path, model_folder_path):
    examples = []
    gold_preds = []

    with open(test_file_path, 'r', encoding='utf-8') as input_file:
        for line in input_file:

            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 2:
                examples.append(parts[0])
                gold_preds.append(int(parts[1]))

    vectorizer = RelationVectorizerBert.from_saved(200, tokenizer_folder_path, True)

    test_data = vectorizer.vectorize(examples)
    model = transformers.TFBertForSequenceClassification.from_pretrained(model_folder_path, num_labels=2)
    prediction = model.predict(test_data)
    prediction_logits = prediction.logits
    probs = tf.nn.softmax(prediction_logits, axis=1).numpy()

    preds = [row[1] for row in probs]

    tp_cnt = 0
    fp_cnt = 0
    tn_cnt = 0
    fn_cnt = 0

    for example, pred, gold_pred in zip(examples, preds, gold_preds):

        if pred > 0.5:
            if gold_pred == 1:
                tp_cnt += 1
            else:
                fp_cnt += 1
                print(f'FP: {example}, {pred}')
        else:
            if gold_pred == 0:
                tn_cnt += 1
            else:
                fn_cnt += 1
                print(f'FN: {example}, {pred}')

    precision = 0

    if tp_cnt + fp_cnt > 0:
        precision = tp_cnt / (tp_cnt + fp_cnt)

    recall = tp_cnt / (tp_cnt + fn_cnt)

    f_score = 0
    if precision + recall > 0:
        f_score = (2 * precision * recall) / (precision + recall)

    print(f'TP: {tp_cnt}, FP: {fp_cnt}, TN: {tn_cnt}, FN: {fn_cnt}, Precision: {precision}, Recall: {recall},'
          f' F-Score: {f_score}')



import re
from typing import List, Tuple
import typing
from datasketch import MinHashLSH, MinHash
from quid.match.Match import Match
from rapidfuzz.distance import Levenshtein
from proquo.core.Quote import Quote
import logging


HASH_PERM: int = 128
LSH_THRESHOLD: float = 0.70

# 1. " und "
# 2. „ und “
# 3. „ und "
# 4. “ und “
# 5. » und «
# 6. « und »
# 7. ‘ und ’
QUOTATION_MARKS = [('"', '"'), ('\u201E', '\u201C'), ('\u201E', '"'), ('\u201C', '\u201C'),
                   ('\u00BB', '\u00AB'), ('\u00AB', '\u00BB'), ('\u2018', '\u2019')]


def filter_short_matches(source_text: str, target_text: str, matches: List[Match], min_length: int) -> List[Match]:
    """
    Takes a list of matches and only returns the ones which are shorter than the given length in words.
    :param source_text: A source text
    :param target_text: A target text
    :param matches: A list of matches
    :param min_length: The maximum length in words
    :return: A list of matches which fulfill the criteria
    """
    result: List[Match] = []
    for match in matches:
        match_source_text = source_text[match.source_span.start:match.source_span.end]
        match_target_text = target_text[match.target_span.start:match.target_span.end]
        source_length = len(match_source_text.split())
        target_length = len(match_target_text.split())

        if source_length < min_length or target_length < min_length:
            result.append(match)

    return result


def get_quotations(input_text: str, max_length: int, quote_open: str = None, quote_close: str = None) -> List[Quote]:
    """
    Search for text in quotation marks which is shorter than the given maximum length in words.
    :param quote_close: The opening quotation mark
    :param quote_open: The closing quotation mark
    :param input_text: The input length
    :param max_length: The maximum length in words
    :return: A list of found quotations
    """

    if not quote_open or not quote_close:
        quote_open, quote_close = __determine_quotation_mark(input_text)

    if not quote_open or not quote_close:
        return []

    quotes: List[Quote] = []

    for re_match in re.finditer(fr'{quote_open}([^{quote_open}]+){quote_close}', input_text):
        start = re_match.span(1)[0]
        end = re_match.span(1)[1]
        text = re_match.group(1)

        if len(text.split()) < max_length:
            quotes.append(Quote(start, end, text))

    return quotes


def __determine_quotation_mark(input_text: str) -> Tuple[str, str]:
    quote_open = ''
    quote_close = ''
    best_count = 0

    for qp in QUOTATION_MARKS:
        match_count = sum(1 for _ in re.finditer(fr'{qp[0]}\b([^{qp[0]}]+){qp[1]}', input_text))

        if match_count > 0:
            open_count = input_text.count(qp[0])
            close_count = input_text.count(qp[1])

            if open_count != 0 and close_count == 0:
                logging.warning(f'Found only left quotation marks for {qp[0]} and {qp[1]}!')
                continue

            if open_count == 0 and close_count != 0:
                logging.warning(f'Found only right quotation marks for {qp[0]} and {qp[1]}!')
                continue

            if open_count != close_count:
                logging.warning(f'Found {match_count} quotes but quote left and right count are not the same:'
                                f' {open_count} vs. {close_count} for {qp[0]} and {qp[1]}')

            if match_count > best_count:
                best_count = match_count
                quote_open = qp[0]
                quote_close = qp[1]

    if not quote_open or not quote_close:
        logging.warning('No quote type could be determined!')
    else:
        logging.info(f'Best quote type: {quote_open} and {quote_close}')

    return quote_open, quote_close


def clean_text(input_text: str) -> str:
    """
    Clean the given text without changing the length, i.e. replace all special characters with spaces.
    :param input_text: The input text
    :return: The cleaned text
    """
    output_text = re.sub('(\\[\\.\\.\\.]|\\[…]|\\.\\.\\.|…)', lambda x: ' ' * len(x.group(1)), input_text)
    output_text = re.sub(f'[^a-zA-Z\\däüöÄÜÖß\n ]', ' ', output_text)

    if len(input_text) != len(output_text):
        raise Exception('Length of source text changed')

    return output_text.lower()


def __clean_single_word(input_text: str) -> str:
    """
    Clean the given text by removing special characters.
    :param input_text: The input text
    :return: The cleaned text
    """
    output_text = re.sub(f'[^a-zA-Z\\däüöÄÜÖß\n ]', '', input_text)
    return output_text.lower()


def normalize_special_chars(input_word: str) -> str:
    """
    Replace some character combinations with another version to normalize umlauts.
    :param input_word:
    :return:
    """
    input_word = input_word.replace('ß', 'ss')
    input_word = input_word.replace('ä', 'ae')
    input_word = input_word.replace('ö', 'oe')
    input_word = input_word.replace('ü', 'ue')
    input_word = input_word.replace('ey', 'ei')

    input_word = input_word.replace('[', '')
    input_word = input_word.replace(']', '')
    return input_word


def init_lsh_hashes(input_text: str) -> MinHashLSH:
    """
    Tokenizes the input text by whitespace, hashes all tokens and creates a MinHashLSH index.
    :param input_text: The input text
    :return: The initialized MinHashLSH index
    """
    hashes = MinHashLSH(threshold=LSH_THRESHOLD, num_perm=HASH_PERM)

    for match in re.finditer(r'\S+', input_text):
        token = normalize_special_chars(match.group())

        if len(token) > 0:
            text_begin_pos = match.start()
            text_end_pos = match.end()

            token_character_set = set(token)
            token_hash = MinHash(num_perm=HASH_PERM)

            for char in token_character_set:
                token_hash.update(char.encode('utf8'))

            hashes.insert(f'{text_begin_pos}_{text_end_pos}_{token}', token_hash, True)

    return hashes


# TODO: improve with better normalization
def strict_match(word: str, search_space: str) -> List[typing.Match[str]]:
    """
    Match the given word against the given text.
    :param word: The word to search for
    :param search_space: The text in which to search
    :return: A list of matches
    """
    word = clean_text(word)
    # word = self.__normalize_special_chars(word)
    # word = re.sub(' +', ' ', word, flags=re.DOTALL)
    # word = word.strip()
    # search_space = self.__normalize_special_chars(search_space)
    re_matches_iter = re.finditer(r'\b' + re.escape(word) + r'\b', search_space, flags=re.IGNORECASE)
    re_matches = list(re_matches_iter)
    return re_matches


def fuzzy_match(word: str, range_start: int, range_end: int, hashes, score_cutoff: float) -> List[Tuple[int, int]]:
    word = __clean_single_word(word)
    word = normalize_special_chars(word)
    word = re.sub(' +', ' ', word, flags=re.DOTALL)
    word = word.strip()
    word_character_set = set(word)
    word_hash = MinHash(num_perm=HASH_PERM)

    for char in word_character_set:
        word_hash.update(char.encode('utf8'))

    candidates = hashes.query(word_hash)
    candidates_split = []

    for candidate in candidates:
        parts = candidate.split('_')
        candidates_split.append((parts[0], parts[1], parts[2]))

    candidates_split.sort(key=lambda e: int(e[0]))

    result = []

    for candidate in candidates_split:
        c_start = int(candidate[0])
        c_end = int(candidate[1])
        token = candidate[2]
        ratio = Levenshtein.normalized_similarity(word, token)

        if ratio < score_cutoff:
            continue

        if c_start >= range_start and c_end <= range_end:
            result.append((c_start, c_end))

    return result


def remove_short_matches(matches: List[Match], target_file_content):
    """
    Remove matches which are shorter than 5 words.
    :return: The remaining matches.
    """

    if len(matches) == 0:
        return []

    filtered_matches = []

    for match in matches:
        text = target_file_content[match.target_span.start:match.target_span.end]
        length = len(text.split())

        if length >= 5:
            filtered_matches.append(match)

    return filtered_matches


def remove_overlapping_matches(matches: List[Match], target_file_content):
    """
    Remove matches which overlap in the target text.
    :param matches: The input list of matches.
    :return: The remaining matches.
    """

    if len(matches) == 0:
        return []

    matches.sort(key=lambda x: x.target_span.start, reverse=False)

    result: List[Match] = []
    match_position: int = 1
    current_match = matches[0]

    while match_position < len(matches):
        next_match = matches[match_position]

        current_target_match_span = current_match.target_span
        next_target_match_span = next_match.target_span

        current_end_pos = current_target_match_span.end
        next_start_pos = next_target_match_span.start

        if next_start_pos >= current_end_pos:
            result.append(current_match)
            current_match = next_match
        else:
            current_match_text = target_file_content[current_target_match_span.start:current_target_match_span.end]
            next_match_text = target_file_content[next_target_match_span.start:next_target_match_span.end]
            current_token_length = len(current_match_text.split())
            next_token_length = len(next_match_text.split())

            if current_token_length < next_token_length:
                current_match = next_match

        match_position += 1

    result.append(current_match)

    return result



from proquo.model.reference.ReferenceModelTrainer import ReferenceModelTrainer
from proquo.model.reference.ReferenceVectorizer import ReferenceVectorizer


def test(input_path, vocab_path, model_path):
    vectorizer = ReferenceVectorizer.from_vocab_file(vocab_path, 25, True)

    sentences_1 = []
    sentences_2 = []
    gold_preds = []

    with open(input_path, 'r', encoding='utf-8') as input_file:
        for line in input_file:
            if not line.strip():
                continue

            parts = line.split('\t')

            if len(parts) == 3:
                sentences_1.append(parts[0])
                sentences_2.append(parts[1])
                gold_preds.append(int(parts[2]))

    test_data_x_1 = vectorizer.vectorize(sentences_1)
    test_data_x_2 = vectorizer.vectorize(sentences_2)

    reference_vectorizer = ReferenceVectorizer.from_vocab_file(vocab_path, 25, True)
    reference_model_trainer = ReferenceModelTrainer(25, True, 32, 32, 0.2, 512, 10)
    reference_model = reference_model_trainer.get_model(reference_vectorizer.max_id)
    reference_model.load_weights(model_path)

    print(reference_model.summary())

    preds = list(reference_model.predict([test_data_x_1, test_data_x_2], verbose=1).ravel())

    tp_cnt = 0
    fp_cnt = 0
    tn_cnt = 0
    fn_cnt = 0

    for s1, s2, pred, gold_pred in zip(sentences_1, sentences_2, preds, gold_preds):
        if pred > 0.5:
            if gold_pred == 1:
                tp_cnt += 1
            else:
                fp_cnt += 1
                print(f'FP: {s1}, {s2}, {pred}')
        else:
            if gold_pred == 0:
                tn_cnt += 1
            else:
                fn_cnt += 1
                print(f'FN: {s1}, {s2}, {pred}')

    precision = tp_cnt / (tp_cnt + fp_cnt)
    recall = tp_cnt / (tp_cnt + fn_cnt)

    print(f'TP: {tp_cnt}, FP: {fp_cnt}, TN: {tn_cnt}, FN: {fn_cnt}, Precision: {precision}, Recall: {recall}')



import tensorflow as tf
import transformers
from proquo.model.linking.LinkingVectorizer import LinkingVectorizer


def test(input_path, tokenizer_path, model_path, lower_case):
    names = []
    examples = []
    gold_preds = []
    quotes = []

    with open(input_path, 'r', encoding='utf-8') as input_file:
        for line in input_file:

            if not line.strip():
                continue

            parts = line.strip().split('\t')

            if len(parts) >= 3:
                examples.append((parts[0], parts[1]))
                gold_preds.append(int(parts[2]))

                if len(parts) >= 6:
                    names.append(parts[3])
                    quotes.append(f'{parts[4]} - {parts[5]}')
                else:
                    names.append('NN')
                    quotes.append('NN')

    link_bert_vectorizer = LinkingVectorizer.from_saved(512, tokenizer_path, lower_case)
    link_bert_model = transformers.TFBertForSequenceClassification.from_pretrained(model_path, num_labels=2)

    test_data = link_bert_vectorizer.vectorize(examples)

    prediction = link_bert_model.predict(test_data)
    prediction_logits = prediction.logits
    probs = tf.nn.softmax(prediction_logits, axis=1).numpy()
    preds = [row[1] for row in probs]

    tp_cnt = 0
    fp_cnt = 0
    tn_cnt = 0
    fn_cnt = 0

    for name, example, quote, pred, gold_pred in zip(names, examples, quotes, preds, gold_preds):

        if pred > 0.5:
            if gold_pred == 1:
                tp_cnt += 1
            else:
                fp_cnt += 1
                print(f'FP: {example}, {pred}, {name}, {quote}')
        else:
            if gold_pred == 0:
                tn_cnt += 1
            else:
                fn_cnt += 1
                print(f'FN: {example}, {pred}, {name}, {quote}')

    precision = tp_cnt / (tp_cnt + fp_cnt)
    recall = tp_cnt / (tp_cnt + fn_cnt)

    f_score = 0
    if precision + recall > 0:
        f_score = (2 * precision * recall) / (precision + recall)

    print(f'TP: {tp_cnt}, FP: {fp_cnt}, TN: {tn_cnt}, FN: {fn_cnt}, Precision: {precision}, Recall: {recall},'
          f' F-Score: {f_score}')



from tensorflow.keras.utils import Sequence
from tensorflow.keras.layers import Dense, Input, LSTM, Bidirectional, Lambda, Embedding
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers.legacy import Adam
import numpy as np
from os.path import join


from proquo.model.reference.ReferenceVectorizer import ReferenceVectorizer


class DataGenerator(Sequence):

    def __init__(self, data_1, data_2, labels, batch_size):
        self.data_1 = data_1
        self.data_2 = data_2
        self.labels = labels
        self.batch_size = batch_size

    def __len__(self):
        return (np.ceil(len(self.data_1) / self.batch_size)).astype(np.int32)

    def __getitem__(self, idx):
        batch_x_1 = self.data_1[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_x_2 = self.data_2[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_y = self.labels[idx * self.batch_size: (idx + 1) * self.batch_size]
        return [batch_x_1, batch_x_2], batch_y


class ReferenceModelTrainer:
    def __init__(self, max_sequence_length, lower_case, num_dense_units, num_lstm_units, dropout_rate, batch_size,
                 num_epochs):
        self.max_sequence_length = max_sequence_length
        self.lower_case = lower_case
        self.num_dense_units = num_dense_units
        self.num_lstm_units = num_lstm_units
        self.dropout_rate = dropout_rate
        self.batch_size = batch_size
        self.num_epochs = num_epochs

    def get_model(self, max_id):
        model_input_1 = Input(shape=(self.max_sequence_length,))
        model_input_2 = Input(shape=(self.max_sequence_length,))

        embedding_layer = Embedding(input_dim=max_id + 1, output_dim=self.num_dense_units,
                                    input_length=self.max_sequence_length, mask_zero=True)

        embedding_1 = embedding_layer(model_input_1)
        embedding_2 = embedding_layer(model_input_2)

        lstm = Bidirectional(LSTM(self.num_lstm_units, return_sequences=False, dropout=self.dropout_rate))

        x1 = lstm(embedding_1)
        x2 = lstm(embedding_2)

        l1_norm = lambda x: 1 - abs(x[0] - x[1])
        merged = Lambda(function=l1_norm, output_shape=lambda x: x[0], name='L1_distance')([x1, x2])
        preds = Dense(1, activation='sigmoid')(merged)
        model = Model(inputs=[model_input_1, model_input_2], outputs=preds)

        return model

    def train_model(self, train_word_pairs, train_labels, val_word_pairs, val_labels, output_dir):
        single_list = []

        for wp in train_word_pairs:
            single_list.append(wp[0])
            single_list.append(wp[1])

        vectorizer = ReferenceVectorizer.from_raw(single_list, self.max_sequence_length, self.lower_case)

        vocab = ''
        for key, value in vectorizer.tokenizer.word_index.items():
            if value == 1:
                vocab += key
            else:
                vocab += f'\n{key}'

        vocab_path = join(output_dir, 'vocab.txt')
        with open(vocab_path, "w", encoding='utf-8') as vocab_file:
            vocab_file.write(vocab)

        train_sentences_1 = [x[0] for x in train_word_pairs]
        train_sentences_2 = [x[1] for x in train_word_pairs]
        val_sentences_1 = [x[0] for x in val_word_pairs]
        val_sentences_2 = [x[1] for x in val_word_pairs]

        train_data_x_1 = vectorizer.vectorize(train_sentences_1)
        train_data_x_2 = vectorizer.vectorize(train_sentences_2)
        val_data_x_1 = vectorizer.vectorize(val_sentences_1)
        val_data_x_2 = vectorizer.vectorize(val_sentences_2)

        train_labels = np.asarray(train_labels).astype(np.float16).reshape((-1, 1))
        val_labels = np.asarray(val_labels).astype(np.float16).reshape((-1, 1))

        model = self.get_model(vectorizer.max_id)
        model.compile(loss='binary_crossentropy', optimizer=Adam(0.001), metrics=['acc'])

        early_stopping = EarlyStopping(monitor='val_loss', patience=3)
        name = f'lstm_{self.num_lstm_units}_{self.dropout_rate}'

        best_model_path = join(output_dir, name + '.h5')
        model_checkpoint = ModelCheckpoint(best_model_path, save_best_only=True, save_weights_only=False)

        print(model.summary())

        training_generator = DataGenerator(train_data_x_1, train_data_x_2, train_labels, self.batch_size)
        validation_generator = DataGenerator(val_data_x_1, val_data_x_2, val_labels, self.batch_size)

        model.fit(x=training_generator,
                  steps_per_epoch=int(len(train_data_x_1) // self.batch_size),
                  epochs=self.num_epochs,
                  validation_data=validation_generator,
                  validation_steps=int(len(val_data_x_1) // self.batch_size),
                  callbacks=[early_stopping, model_checkpoint])



from typing import List
import transformers
import tensorflow as tf


class RelationVectorizerBert:
    max_length: int
    vocab_size: int
    tokenizer: transformers.BertTokenizer

    def __init__(self, max_length, tokenizer):
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.vocab_size = len(self.tokenizer)

    @classmethod
    def from_raw(cls, max_length, lower_case):
        tokenizer = transformers.BertTokenizer.from_pretrained(
            "bert-base-german-dbmdz-uncased", do_lower_case=lower_case
        )

        special_tokens_dict = {'additional_special_tokens': ['<REF>', '<OREF>', '<Q>', '</Q>']}
        num_added_tokens = tokenizer.add_special_tokens(special_tokens_dict)
        print(f'Added {num_added_tokens} tokens')

        return cls(max_length, tokenizer)

    @classmethod
    def from_saved(cls, max_length, tokenizer_path, lower_case):
        tokenizer = transformers.BertTokenizer.from_pretrained(
            tokenizer_path, do_lower_case=lower_case
        )

        return cls(max_length, tokenizer)

    def vectorize(self, input_list: List[str]):
        encoded = self.tokenizer.batch_encode_plus(
            input_list,
            truncation=True,
            add_special_tokens=True,
            max_length=self.max_length,
            return_attention_mask=True,
            return_token_type_ids=True,
            pad_to_max_length=True,
            return_tensors="tf",
        )

        # Convert batch of encoded features to numpy array.
        input_ids = tf.constant(encoded["input_ids"], dtype="int32")
        attention_masks = tf.constant(encoded["attention_mask"], dtype="int32")
        token_type_ids = tf.constant(encoded["token_type_ids"], dtype="int32")

        return input_ids, attention_masks, token_type_ids



import sys
import itertools
import re
from typing import Optional, List, Tuple
from quid.match.Match import Match
from quid.match.MatchSpan import MatchSpan

from proquo.core import Helper
from proquo.match.MatchRef import MatchRef
from proquo.core.Quote import Quote
from proquo.core.QuoteRef import QuoteRef
from proquo.core.Reference import Reference
import numpy as np
from sklearn.cluster import SpectralClustering
import tensorflow as tf
from transformers.modeling_tf_utils import TFPreTrainedModel
from tensorflow.keras.models import Model

from proquo.model.reference.ReferenceVectorizer import ReferenceVectorizer
from proquo.model.relation.RelationVectorizerBert import RelationVectorizerBert

from kpcommons import Footnote, Util


# noinspection PyMethodMayBeStatic
class ProQuo:
    # Todo: Make these customizable
    # References
    MAX_BRACKET_LENGTH = 25
    LONG_MIN_LENGTH = 5
    REF_EXAMPLE_COUNT = 5
    # Relation
    REF_MIN_PROB: float = 0.7
    REL_MIN_PROB: float = 0.5
    REL_MAX_DISTANCE = 100
    REL_MAX_WORDS: int = 120
    # Linking
    WITHOUT_REF_SEARCH_RADIUS = 500
    # Fuzzy matching
    SCORE_CUTOFF: int = 0.85

    def __init__(self, reference_model: Model, reference_vectorizer: ReferenceVectorizer,
                 relation_model: TFPreTrainedModel, relation_vectorizer: RelationVectorizerBert, open_quote: str = None,
                 close_quote: str = None):
        """
        :param reference_model: The model for reference classification
        :param reference_vectorizer: The vectorizer to use to tokenize and encode inputs for the reference model
        :param relation_model: The model for relation classification
        :param relation_vectorizer: The vectorizer to use to tokenize and encode inputs for the relation model
        :param open_quote: Character used for opening quotation marks
        :param close_quote: Character used for closing quotation marks
        """
        self.reference_model = reference_model
        self.reference_vectorizer = reference_vectorizer
        self.relation_model = relation_model
        self.relation_vectorizer = relation_vectorizer
        self.source_cache = {}
        self.hashes = None
        self.parallel_print = False
        self.parallel_print_first_page = 0
        self.parallel_print_last_page = 0
        self.open_quote = open_quote
        self.close_quote = close_quote

    def compare(self, source_text: str, target_text: str, quid_matches: List[Match],
                parallel_print=False, parallel_print_first_page=0, parallel_print_last_page=0)\
            -> List[MatchRef]:
        """
        Compare the two input texts and return a list of matching sequences.
        :param source_text: A source text
        :param target_text: A target text
        :param quid_matches: A list of matches from Quid to be used as candidates for short matches
        :param parallel_print: True if the source text is a parallel print edition
        :param parallel_print_last_page: The first page on which the parallel print edition starts
        :param parallel_print_first_page: The last page on which the parallel print edition appears
        :return: A list of matches
        """

        source_text = Helper.clean_text(source_text)

        # Check if text was already processed and use cached version for speed up
        source_text_hash = hash(source_text)
        if source_text_hash in self.source_cache:
            self.hashes = self.source_cache[source_text_hash]
        else:
            self.hashes = Helper.init_lsh_hashes(source_text)
            self.source_cache[source_text_hash] = self.hashes

        self.parallel_print = parallel_print
        self.parallel_print_first_page = parallel_print_first_page
        self.parallel_print_last_page = parallel_print_last_page

        # sort Quid matches into two lists: long and short matches
        long_matches = self.__filter_long_matches(source_text, quid_matches)
        short_matches = Helper.filter_short_matches(source_text, target_text, quid_matches, self.LONG_MIN_LENGTH)

        matches_with_reference: List[MatchRef] = self.__add_reference_to_match(target_text, long_matches)
        matches_with_reference.sort(key=lambda x: x.target_span.end - x.target_span.start, reverse=True)

        # find the n longest matches with a reference
        longest_matches: List[MatchRef] = []
        for match_with_reference in matches_with_reference:
            if match_with_reference.reference:
                if re.search(r'\d', match_with_reference.reference.text):
                    longest_matches.append(match_with_reference)

            if len(longest_matches) >= self.REF_EXAMPLE_COUNT:
                break

        prototypical_reference = self.__get_prototypical_reference(longest_matches)
        all_references: List[Reference] = self.__get_references(target_text, prototypical_reference)
        all_quotes: List[Quote] = Helper.get_quotations(target_text, self.LONG_MIN_LENGTH, self.open_quote,
                                                        self.close_quote)
        footnote_ranges = Footnote.get_footnote_ranges_without_offset(target_text)

        # we are only interested in quotes and references in the running text
        main_references = []
        for ref in all_references:
            if not Footnote.is_range_in_ranges(ref.start, ref.end, footnote_ranges):
                main_references.append(ref)

        main_quotes = []
        for q in all_quotes:
            if not Footnote.is_range_in_ranges(q.start, q.end, footnote_ranges):
                main_quotes.append(q)

        main_quotes = self.__filter_quotes_not_in_source(main_quotes, 0, len(source_text))

        main_references.sort(key=lambda x: x.start)
        main_quotes.sort(key=lambda x: x.start)

        quote_ref_combos_main = self.__get_combinations(target_text, main_references, main_quotes, footnote_ranges)

        quote_ref_combos: List[QuoteRef] = []
        quote_ref_combos.extend(quote_ref_combos_main)

        short_quotes: List[QuoteRef] = self.__get_short_quotes(quote_ref_combos)
        short_quotes = self.__filter_duplicates(short_quotes)

        matches_with_reference.sort(key=lambda x: 0 if x.reference is None else x.reference.page)
        result_matches: List[MatchRef] = []

        page_size, first_page_known_pos, first_page_nr = self.__get_page_size(matches_with_reference)

        # quotes with a reference are processed first
        for sq in short_quotes:
            start, end, page_known_pos = self.__get_quote_range_with_ref(sq, page_size, first_page_known_pos,
                                                                         first_page_nr, len(source_text))

            match_len = len(sq.quote.text.split())

            if match_len == 1:
                quote_source_start, quote_source_end = self.__search_single_word(sq.quote, source_text, start, end,
                                                                                 page_known_pos)
            else:
                quote_source_start, quote_source_end = self.__search_multi_word(short_matches, sq, start, end,
                                                                                source_text)

            if quote_source_start > -1:
                quote_source_text = source_text[quote_source_start:quote_source_end]
                source_span = MatchSpan(quote_source_start, quote_source_end, quote_source_text)
                target_span = MatchSpan(sq.quote.start, sq.quote.end, sq.quote.text)
                match = MatchRef(source_span, target_span, sq.reference)
                result_matches.append(match)

        # then quotes without a reference are processed
        short_quotes_without_ref = []

        for mq in main_quotes:
            found = False

            for sq in short_quotes:
                if sq.quote.start == mq.start and sq.quote.end == mq.end:
                    found = True
                    break

            if not found:
                short_quotes_without_ref.append(QuoteRef(mq, None, ''))

        # add result matches
        matches_with_reference.extend(result_matches)
        matches_with_ref_in_appearance = sorted(matches_with_reference, key=lambda x: x.target_span.start)

        # search for short quotes without a reference
        for sq in short_quotes_without_ref:
            start, end, page_known_pos = self.__get_quote_range_without_ref(sq, matches_with_ref_in_appearance,
                                                                            len(source_text), page_size)

            quote_source_start = -1
            quote_source_end = -1
            match_len = len(sq.quote.text.split())

            if match_len == 1:
                if page_known_pos > -1:
                    quote_source_start, quote_source_end = self.__search_single_word(sq.quote, source_text, start, end,
                                                                                     page_known_pos)
            else:
                quote_source_start, quote_source_end = self.__search_multi_word(short_matches, sq, start, end,
                                                                                source_text)

            if quote_source_start > -1:
                quote_source_text = source_text[quote_source_start:quote_source_end]

                source_span = MatchSpan(quote_source_start, quote_source_end, quote_source_text)
                target_span = MatchSpan(sq.quote.start, sq.quote.end, sq.quote.text)
                match = MatchRef(source_span, target_span, sq.reference)
                result_matches.append(match)

        return result_matches

    def __predict_ref(self, references_1: List[str], references_2: List[str]) -> List[float]:
        """
        Takes two lists of reference strings and returns the probabilities of two references having the same style.
        :param references_1: List of reference strings
        :param references_2: List of references strings
        :return: Probabilities of two references having the same style
        """
        if len(references_1) == 0 or len(references_2) == 0:
            return []

        test_data_x_1 = self.reference_vectorizer.vectorize(references_1)
        test_data_x_2 = self.reference_vectorizer.vectorize(references_2)
        preds = list(self.reference_model.predict([test_data_x_1, test_data_x_2], verbose=0).ravel())
        return preds

    def __predict_rel(self, quote_ref_combos: List[QuoteRef]) -> List[float]:
        """
        Takes a list of combinations of quotation and reference and predicts the probability for each combination that
        the quotation and reference belong together.
        :param quote_ref_combos: A list of combinations of quotation and reference
        :return: A list of probabilities
        """

        if len(quote_ref_combos) == 0:
            return []

        sentences = []

        for qrc in quote_ref_combos:
            text = qrc.text
            text = re.sub(fr'([{self.open_quote}{self.close_quote}])', r' \1 ', text, flags=re.DOTALL)
            sentences.append(text)

        test_data = self.relation_vectorizer.vectorize(sentences)

        prediction = self.relation_model.predict(test_data, verbose=0)
        prediction_logits = prediction.logits
        probs = tf.nn.softmax(prediction_logits, axis=1).numpy()
        preds = [row[1] for row in probs]

        return preds

    def __filter_long_matches(self, input_text: str, matches: List[Match]) -> List[Match]:
        """
        Takes a list of matches and returns a new list which only contains matches which are longer then a certain
        lower limit.
        :param input_text: The input text
        :param matches: A list of matches
        :return: A list of matches which fulfill the length requirement
        """
        result = []
        for match in matches:
            text = input_text[match.source_span.start:match.source_span.end]
            length = len(text.split())

            if length >= self.LONG_MIN_LENGTH:
                result.append(match)

        return result

    def __add_reference_to_match(self, input_text: str, matches: List[Match]) -> List[MatchRef]:
        """
        Goes through all matches and tries to find a reference in parentheses after the match within a certain distance.
        :param input_text: A text
        :param matches: A list of matches
        :return: A list of matches with references
        """

        result: List[MatchRef] = []

        for match in matches:
            match_ref = MatchRef.from_match(match)
            end = match.target_span.end
            context_text = input_text[end:end + 20]

            re_match = re.search(r'\((.+?)\)', context_text)
            if re_match:
                ref_start = re_match.start(1)
                ref_end = re_match.end(1)
                ref_text = re_match.group(1)
                page_nr = self.__extract_page_from_reference(ref_text)

                reference = Reference(ref_start, ref_end, ref_text, page_nr)
                match_ref.reference = reference

            result.append(match_ref)

        return result

    def __get_prototypical_reference(self, match_ref_examples: List[MatchRef]) -> str:
        """
        From the up to five examples, the best example is identified.
        :param match_ref_examples: A list of matches with references
        :return: A prototypical reference string.
        """
        num_examples = len(match_ref_examples)

        if num_examples == 0:
            return ''

        match_ref_ex_with_id = []

        # as a default, we just use the first example
        index = 0

        # To identify the best example, we need more than two examples
        if num_examples > 2:
            for pos, mre in enumerate(match_ref_examples):
                match_ref_ex_with_id.append((pos, mre))

            match_ref_combinations = itertools.combinations(match_ref_ex_with_id, 2)
            sentences_1: List[str] = []
            sentences_2: List[str] = []
            combinations = []

            perm: Tuple[Tuple[int, MatchRef], Tuple[int, MatchRef]]
            for perm in match_ref_combinations:
                sentences_1.append(perm[0][1].reference.text)
                sentences_2.append(perm[1][1].reference.text)
                combinations.append((perm[0], perm[1]))

            if len(sentences_1) == 0 or len(sentences_2) == 0:
                return ''

            preds = self.__predict_ref(sentences_1, sentences_2)
            num_examples = len(match_ref_examples)

            matrix = np.eye(num_examples, dtype=float)

            for perm, pred in zip(combinations, preds):
                x = perm[0][0]
                y = perm[1][0]
                # matrix has to be symmetrical
                matrix[x][y] = pred
                matrix[y][x] = pred

            labels = SpectralClustering(2, affinity='precomputed').fit_predict(matrix)
            labels = labels.tolist()

            num_occ_cluster_0 = labels.count(0)
            num_occ_cluster_1 = labels.count(1)

            if num_occ_cluster_0 > num_occ_cluster_1:
                index = labels.index(0)
            else:
                index = labels.index(1)

        best_bibl = match_ref_examples[index]
        return best_bibl.reference.text

    def __get_references(self, input_text: str, prototypical_reference: str) -> List[Reference]:
        """
        All text in parentheses is compared to the prototypical example to identify all references in the text.
        :param input_text: A text
        :param prototypical_reference: A prototypical reference string
        :return: A list of all references in the text which are similar to the given examples.
        """

        if not prototypical_reference:
            return []

        bracket_positions = {}
        stack = []

        for i, c in enumerate(input_text):
            if c == '(':
                stack.append(i)
            elif c == ')':
                if len(stack) > 0:
                    # There will be errors, so we just ignore them
                    bracket_positions[stack.pop()] = i

        possible_references = []

        for start, end in bracket_positions.items():
            if end - start > self.MAX_BRACKET_LENGTH:
                continue

            ref_text = input_text[start + 1:end]
            # TODO: resolve ebd.
            page_nr = self.__extract_page_from_reference(ref_text)
            possible_references.append(Reference(start + 1, end, input_text[start + 1:end], page_nr))

        possible_permutations = itertools.product([x.text for x in possible_references], [prototypical_reference])

        poss_bibl_sentences_1 = []
        poss_bibl_sentences_2 = []

        for perm in list(possible_permutations):
            poss_bibl_sentences_1.append(perm[0])
            poss_bibl_sentences_2.append(perm[1])

        preds_2 = self.__predict_ref(poss_bibl_sentences_1, poss_bibl_sentences_2)

        references: List[Reference] = []

        for pr, pred in zip(possible_references, preds_2):
            if pred >= self.REF_MIN_PROB:
                references.append(pr)

        return references

    def __replace_other_ref(self, input_text: str, references: List[Reference], start: int, end: int) -> str:
        """
        Replace reference strings in the given text with a tag in the range between start and end.
        :param input_text: A text
        :param references: A list of references
        :param start: A start character position
        :param end: An end character position
        :return: The modified input text
        """
        other_refs = []

        for ref in references:
            if ref.start >= start and ref.end <= end:
                other_refs.append(ref)

        text = input_text[start:end]
        other_refs.sort(key=lambda x: x.start, reverse=True)

        for ref in other_refs:
            text = f'{text[:ref.start - start]} <OREF> {text[ref.end - start:]}'

        text = re.sub(r'\[\[\[((?:.|\n)+?)]]]', '', text, flags=re.DOTALL)

        return text

    def __get_combinations(self, input_text: str, references: List[Reference], quotes: List[Quote],
                           footnote_ranges: Optional[List[Tuple[int, int]]]) -> List[QuoteRef]:
        """
        Generate all combinations of quotes and reference within a certain distance, excluding footnotes.
        :param input_text: The input text
        :param references: A list of references
        :param quotes: A list of quotes
        :param footnote_ranges: A list of footnote ranges
        :return: A list of combinations of quotes and references. Additionally, a text is created with the quote tagged
        and the reference replaced with a tag.
        """

        my_permutations = itertools.product(quotes, references)

        result: List[QuoteRef] = []

        for perm in my_permutations:
            quote = perm[0]
            reference = perm[1]

            if quote.start < reference.end:
                start = quote.start
                end = reference.end
                between_start = quote.end
                between_end = reference.start
                words_in_fn_count = 0

                if footnote_ranges:
                    words_in_fn_count = self.__count_words_in_footnotes(input_text, footnote_ranges, start, end)

                dist = len(input_text[between_start:between_end].split()) - words_in_fn_count
                if dist > self.REL_MAX_DISTANCE:
                    continue

                text_replaced = self.__replace_other_ref(input_text, references, quote.end, reference.start)
                text = f' <Q> {input_text[quote.start:quote.end]} </Q> {text_replaced[1:]} <REF> '
            else:
                start = reference.start
                end = quote.end
                between_start = reference.end
                between_end = quote.start
                words_in_fn_count = 0

                if footnote_ranges:
                    words_in_fn_count = self.__count_words_in_footnotes(input_text, footnote_ranges, start, end)

                dist = len(input_text[between_start:between_end].split()) - words_in_fn_count
                if dist > self.REL_MAX_DISTANCE:
                    continue

                text_replaced = self.__replace_other_ref(input_text, references, reference.end, quote.start)
                text = f' <REF> {text_replaced[:-1]} <Q> {input_text[quote.start:quote.end]} </Q> '

            length = len(text.split())

            if length > self.REL_MAX_WORDS:
                continue

            rest_len = self.REL_MAX_WORDS - length

            parts_before = input_text[:start].split()
            parts_after = input_text[end:].split()

            parts_before_count = len(parts_before)
            parts_after_count = len(parts_after)

            count_before = min(round(rest_len / 2), parts_before_count)
            count_after = min(rest_len - count_before, parts_after_count)

            text_before = ''
            temp_end = start

            for i in range(len(parts_before) - 1, len(parts_before) - 1 - count_before, -1):
                part = parts_before[i]
                part_len = len(part)
                temp_start = temp_end - 1 - part_len

                if not footnote_ranges or not Footnote.is_range_in_ranges(temp_start, temp_end, footnote_ranges):
                    text_before = f'{parts_before[i]} {text_before}'

                temp_end = temp_start

            text_after = ''
            temp_start = end

            for i in range(0, count_after):
                part = parts_after[i]
                part_len = len(part)
                temp_end = temp_start + 1 + part_len

                if not footnote_ranges or not Footnote.is_range_in_ranges(temp_start, temp_end, footnote_ranges):
                    text_after += f' {parts_after[i]}'

                temp_start = temp_end

            text_before_replaced = self.__replace_other_ref(input_text, references, start - len(text_before), start)
            text_after_replaced = self.__replace_other_ref(input_text, references, end, end + len(text_after))

            if quote.start < reference.end:
                text = f'{text_before_replaced[:-1]} {text[1:]} {text_after_replaced}'
            else:
                text = f'{text_before_replaced} {text[:-1]} {text_after_replaced[1:]}'

            result.append(QuoteRef(quote, reference, text))

        return result

    def __get_short_quotes(self, quote_ref_combos: List[QuoteRef]) -> List[QuoteRef]:
        """
        Takes a list of combinations of quotation and reference and returns a new list which only contains combinations
        which are have a certain probability of belonging together.
        :param quote_ref_combos: A list of combinations of quotation and reference
        :return: A filtered list of combinations of quotation and reference
        """
        preds = self.__predict_rel(quote_ref_combos)
        result = []

        for qrc, pred in zip(quote_ref_combos, preds):
            if pred >= self.REL_MIN_PROB:
                qrc.pred = pred
                result.append(qrc)

        return result

    def __filter_quotes_not_in_source(self, quotes: List[Quote], start: int, end: int) -> List[Quote]:
        """
        Create a list of quotes from the given list of quotes which only contains quotes which are longer than a single
        word or could be found the source text by fuzzing matching.
        :param quotes: A list of quotes
        :param start: A start character position
        :param end: A end character position
        :return: A list of the remaining quotes
        """
        result: List[Quote] = []

        for quote in quotes:
            if len(quote.text.split()) > 1:
                result.append(quote)
                continue

            candidates = Helper.fuzzy_match(quote.text, start, end, self.hashes, self.SCORE_CUTOFF)
            if len(candidates) == 0:
                continue

            result.append(quote)

        return result

    def __filter_duplicates(self, short_quotes: List[QuoteRef]) -> List[QuoteRef]:
        """
        Create a list only containing non overlapping quotes.
        :param short_quotes: The input list of quotes
        :return: The resulting list of quotes
        """
        result = []

        for sq in short_quotes:
            found = False
            replace_pos = -1

            for pos, r in enumerate(result):
                if r.quote.start == sq.quote.start and r.quote.end == sq.quote.end:
                    found = True

                    if sq.pred > r.pred:
                        replace_pos = pos
                    break

            if not found:
                result.append(sq)
            elif replace_pos > -1:
                result[replace_pos] = sq

        return result

    def __calc_page_offset_parallel_print(self, start: int, end: int) -> int:
        """
        Only needed for parallel print editions.
        :param start: A start page
        :param end: An end page
        :return: The number of pages to subtract
        """
        if end > start:
            if self.parallel_print_first_page < end <= self.parallel_print_last_page:
                return (end - max(self.parallel_print_first_page, start)) // 2
            elif self.parallel_print_first_page < start < self.parallel_print_last_page:
                return (self.parallel_print_last_page - start) // 2
        else:
            if self.parallel_print_first_page < start < self.parallel_print_last_page:
                return (start - max(self.parallel_print_first_page, end)) // 2
            elif self.parallel_print_first_page < end < self.parallel_print_last_page:
                return (self.parallel_print_last_page - end) // 2

        return 0

    def __extract_page_from_reference(self, text: str) -> int:
        """
        A simple approach to extracting a page number from a reference string.
        :param text: A reference string
        :return: The extracted page number
        """

        page_nr = 0

        re_match = re.search(r'S\. ?(\d+)', text)
        if re_match:
            page_nr = int(re_match.group(1))
            return page_nr

        re_match = re.search(r'^(\d+)', text)
        if re_match:
            page_nr = int(re_match.group(1))
            return page_nr

        re_match = re.search(r'\W(\d+)', text)
        if re_match:
            page_nr = int(re_match.group(1))
            return page_nr

        # page_nr = self.__resolve_non_number_page(text, prev_match_with_ref)
        return page_nr

    # def __resolve_non_number_page(self, text: str, prev_match_with_ref: MatchRef) -> int:
    #     if not prev_match_with_ref:
    #         return 0
    #
    #     if text.startswith('ebd'):
    #         return prev_match_with_ref.reference.page
    #
    #     return 0

    def __count_words_in_footnotes(self, input_text: str, footnote_ranges: List[Tuple[int, int]], start: int, end: int)\
            -> int:
        """
        Count the number of words in footnotes in the given range.
        :param input_text: The input text
        :param footnote_ranges: The list of footnote ranges
        :param start: A start character position
        :param end: An end character position
        :return: The number of words in footnotes
        """

        count = 0

        for fr in footnote_ranges:
            if start <= fr[0] < end:
                text = input_text[fr[0]:fr[1]]
                count += len(text.split())

        return count

    def __get_page_size(self, matches_with_reference: List[MatchRef]) -> Tuple[int, int, int]:
        """
        Estimate the page size of the source text from the long matches and their references (from Quid).
        This is done by using the quotes which appear first and last in the source text and estimate a page size
        based on the distance of these two quotes and the page numbers.
        :param matches_with_reference: A list of matches and their references
        :return: A tuple with the page size, the character start position of the first quote and the page number
        """
        mwr_sorted = sorted(matches_with_reference, key=lambda x: x.source_span.start)

        start = -1
        end = -1
        first_page = 0
        last_page = 0

        for match_ref in mwr_sorted:
            if match_ref.reference and match_ref.reference.page > 0:
                if start == -1:
                    start = match_ref.source_span.start
                    first_page = match_ref.reference.page
                elif first_page > match_ref.reference.page:
                    start = -1

        for match_ref in reversed(mwr_sorted):
            if match_ref.reference and match_ref.reference.page > 0:
                if end == -1:
                    end = match_ref.source_span.end
                    last_page = match_ref.reference.page
                elif last_page < match_ref.reference.page:
                    end = -1

        page_size = 0

        if self.parallel_print:
            page_diff = last_page - first_page - self.__calc_page_offset_parallel_print(first_page, last_page)
        else:
            page_diff = last_page - first_page

        if 0 < first_page != last_page > 0:
            page_size = ((end - start) // page_diff)

        return page_size, start, first_page

    def __get_quote_range_with_ref(self, sq: QuoteRef, page_size: int, first_page_known_pos: int, first_page_nr: int,
                                   source_text_len: int):
        """
        Tries to estimate the best range in the source text to search for the quotation.
        :param sq: A quote with a reference
        :param page_size: The page size
        :param first_page_known_pos: The character position of the first match
        :param first_page_nr: The page number of the first page
        :param source_text_len: Length of the source text
        :return: A tuple of: The character start and end positions of the range in the source text in which we assume
        the quote can be found. And the estimated position of where the quote is assumed to be found.
        """
        start = 0
        end = source_text_len
        page_known_pos = -1

        if sq.reference.page > 0:
            if page_size > 0 and first_page_nr > 0:

                if self.parallel_print:
                    page_diff = sq.reference.page - first_page_nr - \
                                self.__calc_page_offset_parallel_print(first_page_nr, sq.reference.page)
                else:
                    page_diff = sq.reference.page - first_page_nr

                page_known_pos = first_page_known_pos + (page_diff * page_size)
                start = max(0, page_known_pos - page_size)
                end = min(source_text_len, page_known_pos + page_size)

        # in this special case we search the whole text
        if end <= start:
            start = 0
            end = source_text_len

        return start, end, page_known_pos

    def __get_quote_range_without_ref(self, sq: QuoteRef, matches_with_reference: List[MatchRef], source_text_len: int,
                                      page_size: int) -> Tuple[int, int, int]:
        """
        Tries to estimate the best range in the source text to search for the quotation.
        :param sq: A quote without a references
        :param matches_with_reference: Matches with a reference
        :param source_text_len: The length of the source text
        :param page_size: The page size
        :return: A tuple of: The character start and end positions of the range in which we assume the quote can be
        found. And the estimated position of where the quote is assumed to be found.
        """
        mwr_best = None
        best_dist = sys.maxsize

        for pos, match_with_reference in enumerate(matches_with_reference):
            if not match_with_reference.reference:
                continue

            mwr_start = max(0, match_with_reference.target_span.start - self.WITHOUT_REF_SEARCH_RADIUS)
            mwr_end = min(source_text_len, match_with_reference.target_span.end + self.WITHOUT_REF_SEARCH_RADIUS)

            new_dist = sys.maxsize

            if mwr_start <= sq.quote.start <= match_with_reference.target_span.start:
                new_dist = match_with_reference.target_span.start - sq.quote.end
            elif match_with_reference.target_span.end <= sq.quote.start <= mwr_end:
                new_dist = sq.quote.start - match_with_reference.target_span.end

            if new_dist < best_dist:
                best_dist = new_dist
                mwr_best = match_with_reference

        start = 0
        end = source_text_len
        page_known_pos = -1

        if mwr_best:
            if page_size > 0:
                start = max(0, mwr_best.source_span.start - page_size)
                end = min(source_text_len, mwr_best.source_span.end + page_size)
                page_known_pos = mwr_best.source_span.start + ((mwr_best.source_span.end - mwr_best.source_span.start) // 2)

        return start, end, page_known_pos

    def __search_single_word(self, quote: Quote, source_text: str, range_start: int, range_end: int,
                             page_known_pos: int) -> Tuple[int, int]:
        """
        Search for the given quote in the given source text.
        :param quote: A quote
        :param source_text: The source text
        :param range_start: The character start position of the range to search in
        :param range_end: The character end position of the range to search in
        :param page_known_pos:
        :return: A tuple consisting of the start and end character positions if a match was found. Otherwise (-1, -1)
        """

        source_search_space = source_text[range_start:range_end]
        re_matches = Helper.strict_match(quote.text, source_search_space)
        strict_matches_count = len(re_matches)

        if strict_matches_count == 0:
            fuzzy_candidates = Helper.fuzzy_match(quote.text, range_start, range_end, self.hashes,
                                                  self.SCORE_CUTOFF)

            if len(fuzzy_candidates) == 1:
                return fuzzy_candidates[0][0], fuzzy_candidates[0][1]
            elif len(fuzzy_candidates) > 1:
                smallest_dist = sys.maxsize
                best_match_start_pos = 0
                best_match_end_pos = 0

                for fc in fuzzy_candidates:
                    # TODO: Is start ideal? Maybe compare start and end
                    m_start = fc[0]
                    diff = abs(page_known_pos - m_start)
                    if diff < smallest_dist:
                        smallest_dist = diff
                        best_match_start_pos = m_start
                        best_match_end_pos = fc[1]

                return best_match_start_pos, best_match_end_pos

        elif strict_matches_count == 1:
            re_match = re_matches[0]

            if re_match:
                return re_match.start() + range_start, re_match.end() + range_start

        elif page_known_pos > -1:
            smallest_dist = sys.maxsize
            best_match_start_pos = -1
            best_match_end_pos = -1

            for m in re_matches:
                # TODO: Is start ideal? Maybe compare start and end
                m_start = m.start() + range_start
                diff = abs(page_known_pos - m_start)
                if diff < smallest_dist:
                    smallest_dist = diff
                    best_match_start_pos = m_start
                    best_match_end_pos = m.end() + range_start

            return best_match_start_pos, best_match_end_pos

        return -1, -1

    def __search_multi_word(self, short_matches: List[Match], sq: QuoteRef, range_start: int, range_end: int,
                            source_text: str) -> Tuple[int, int]:
        """
        Search for the given quote in the given source text. First for an exact match and then for a fuzzy match.
        :param short_matches: The matches from Quid which are used as candidates
        :param sq: A quote
        :param range_start: The character start position of the range to search in
        :param range_end: The character end position of the range to search in
        :param source_text: The source text
        :return: A tuple consisting of the start and end character positions if a match was found. Otherwise (-1, -1)
        """

        search_space = source_text[range_start:range_end]
        re_matches = Helper.strict_match(sq.quote.text, search_space)

        if len(re_matches) > 1:
            return -1, -1

        if len(re_matches) == 1:
            return re_matches[0].start() + range_start, re_matches[0].end() + range_start
        elif len(re_matches) == 0:
            re_matches = Helper.strict_match(sq.quote.text, source_text)
            if len(re_matches) > 1:
                return -1, -1

            if len(re_matches) == 1:
                return re_matches[0].start(), re_matches[0].end()

        candidates = []

        for short_match in short_matches:
            overlap_length = Util.calculate_overlap(short_match.target_span.start, short_match.target_span.end,
                                                    sq.quote.start, sq.quote.end)
            quote_length = sq.quote.end - sq.quote.start
            percentage = overlap_length / quote_length

            if percentage >= 0.7:
                candidates.append(short_match)

        candidates_count = len(candidates)

        if candidates_count == 0:
            return -1, -1
        elif candidates_count == 1:
            return candidates[0].source_span.start, candidates[0].source_span.end
        else:
            filtered_candidates = []

            for c in candidates:
                if c.source_span.start >= range_start and c.source_span.end <= range_end:
                    filtered_candidates.append(c)

            if len(filtered_candidates) == 1:
                return filtered_candidates[0].source_span.start, filtered_candidates[0].source_span.end

        return -1, -1
